# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#
import gettext

from typing import Callable, Dict, List, Tuple, Optional

from PyQt6.QtGui import QAction, QIcon, QKeySequence

from PyQt6.QtWidgets import (
    QDialog,
    QFileDialog,
    QHBoxLayout,
    QLabel,
    QMainWindow,
    QMenu,
    QMessageBox,
    QProgressDialog,
    QSizePolicy,
    QStackedLayout,
    QStyleFactory,
    QVBoxLayout,
    QWidget,
)

from PyQt6.QtCore import QObject, QRunnable, QThreadPool, QTimer, Qt, pyqtSignal

from eyegrade import exams

from .. import utils
from .. import scoring
from ..stats import stats_core
from . import examsview
from . import widgets
from . import wizards
from . import dialogs
from . import export
from . import students
from . import FileNameFilters
from . import scores
from . import show_stats

t = gettext.translation("eyegrade", utils.locale_dir(), fallback=True)
_ = t.gettext


class _WorkerSignalEmitter(QObject):
    """Convenience class for generating signals from a Worker."""

    finished = pyqtSignal()


class Worker(QRunnable):
    """Generic worker class for spawning a task to other thread."""

    _active_workers: List["Worker"] = []
    _worker_count = 0

    def __init__(self, task):
        """Inits a new worker.

        The `task` must be an object that implements a `run()` method.

        """
        super().__init__()
        self.task = task
        self.is_done = False
        self.signals = _WorkerSignalEmitter()
        if Worker._worker_count > 63:
            Worker._cleanup_done_workers()
        Worker._active_workers.append(self)
        Worker._worker_count += 1

    def run(self):
        """Run the task and emit the signal at its completion."""
        self.task.run()
        self.is_done = True
        self.signals.finished.emit()

    @property
    def finished(self):
        """The `finished` signal as a property."""
        return self.signals.finished

    @staticmethod
    def _cleanup_done_workers():
        Worker._active_workers = [w for w in Worker._active_workers if not w.is_done]
        Worker._worker_count = len(Worker._active_workers)


class ActionsManager:
    """Creates and manages the toolbar buttons."""

    _actions_grading_data: List[Tuple[str, Optional[str], Optional[str], List[int]]] = [
        ("start", "start.svg", _("&Start grading"), []),
        ("stop", "stop.svg", _("S&top grading"), []),
        ("back", "back.svg", _("&Back to session home"), []),
        (
            "continue",
            "continue.svg",
            _("Continue to the &next exam"),
            [Qt.Key.Key_Space],
        ),
        ("*separator*", None, None, []),
        ("snapshot", "snapshot.svg", _("&Capture the current image"), [Qt.Key.Key_C]),
        (
            "manual_detect",
            "manual_detect.svg",
            _("&Manual detection of answer tables"),
            [Qt.Key.Key_M],
        ),
        ("edit_id", "edit_id.svg", _("&Edit student id"), [Qt.Key.Key_I]),
        (
            "discard",
            "discard.svg",
            _("&Discard exam"),
            [Qt.Key.Key_Delete, Qt.Key.Key_Backspace],
        ),
    ]

    _actions_session_data: List[Tuple[str, Optional[str], Optional[str], List[int]]] = [
        ("new", "new.svg", _("&New session"), []),
        ("open", "open.svg", _("&Open session"), []),
        ("close", "close.svg", _("&Close session"), [Qt.Key.Key_Escape]),
        ("*separator*", None, None, []),
        ("exit", "exit.svg", _("&Exit"), []),
    ]

    _actions_exams_data: List[Tuple[str, Optional[str], Optional[str], List[int]]] = [
        ("search", "search.svg", _("&Search"), []),
        ("students", None, _("S&tudents"), []),
        ("export", "export.svg", _("&Export grades listing"), []),
        ("statistics", None, _("S&how statistics"), []),
    ]

    _actions_tools_data: List[Tuple[str, Optional[str], Optional[str], List[int]]] = [
        ("camera", "camera.svg", _("Select &camera"), []),
        ("export_exam_config", None, _("E&xport exam configuration"), []),
        ("edit_scores", None, _("Edit question &scores"), []),
        ("+auto_change", None, _("Continue on exam &removal"), []),
    ]

    _actions_help_data: List[Tuple[str, Optional[str], Optional[str], List[int]]] = [
        ("help", None, _("Online &Help"), []),
        ("website", None, _("&Website"), []),
        ("source", None, _("&Source code at GitHub"), []),
        ("about", None, _("&About"), []),
    ]

    _actions_debug_data: List[Tuple[str, Optional[str], Optional[str], List[int]]] = [
        ("+show_status", None, _("Show &status"), []),
        ("+lines", None, _("Show &lines"), []),
        ("+processed", None, _("Show &processed image"), []),
    ]

    _actions_experimental: List[Tuple[str, Optional[str], Optional[str], List[int]]] = (
        []
    )

    def __init__(self, window):
        """Creates a manager for the given toolbar object."""
        self.window = window
        self.menubar = window.menuBar()
        self.toolbar = window.addToolBar("Grade Toolbar")
        self.menus = {}
        self.actions_grading = {}
        self.actions_session = {}
        self.actions_exams = {}
        self.actions_tools = {}
        self.actions_help = {}
        action_lists = {
            "session": [],
            "grading": [],
            "exams": [],
            "tools": [],
            "help": [],
        }
        for key, icon, text, shortcuts in ActionsManager._actions_session_data:
            self._add_action(
                key,
                icon,
                text,
                shortcuts,
                self.actions_session,
                action_lists["session"],
            )
        for key, icon, text, shortcuts in ActionsManager._actions_grading_data:
            self._add_action(
                key,
                icon,
                text,
                shortcuts,
                self.actions_grading,
                action_lists["grading"],
            )
        for key, icon, text, shortcuts in ActionsManager._actions_exams_data:
            self._add_action(
                key, icon, text, shortcuts, self.actions_exams, action_lists["exams"]
            )
        for key, icon, text, shortcuts in ActionsManager._actions_tools_data:
            self._add_action(
                key, icon, text, shortcuts, self.actions_tools, action_lists["tools"]
            )
        for key, icon, text, shortcuts in ActionsManager._actions_help_data:
            self._add_action(
                key, icon, text, shortcuts, self.actions_help, action_lists["help"]
            )
        self._populate_menubar(action_lists)
        self._populate_toolbar(action_lists)
        self._add_debug_actions()
        self._add_experimental_actions()

    def set_search_mode(self):
        self.actions_grading["start"].setEnabled(False)
        self.actions_grading["stop"].setEnabled(True)
        self.actions_grading["back"].setEnabled(False)
        self.actions_grading["snapshot"].setEnabled(True)
        self.actions_grading["manual_detect"].setEnabled(True)
        self.actions_grading["edit_id"].setEnabled(False)
        self.actions_grading["continue"].setEnabled(False)
        self.actions_grading["discard"].setEnabled(False)
        self.actions_session["new"].setEnabled(False)
        self.actions_session["open"].setEnabled(False)
        self.actions_session["close"].setEnabled(True)
        self.actions_session["exit"].setEnabled(True)
        self.actions_tools["camera"].setEnabled(False)
        self.actions_tools["export_exam_config"].setEnabled(True)
        self.actions_tools["edit_scores"].setEnabled(True)
        self.actions_exams["search"].setEnabled(False)
        self.actions_exams["students"].setEnabled(False)
        self.actions_exams["export"].setEnabled(False)
        self.actions_exams["statistics"].setEnabled(False)

    def set_review_from_grading_mode(self):
        self.actions_grading["start"].setEnabled(False)
        self.actions_grading["stop"].setEnabled(True)
        self.actions_grading["back"].setEnabled(False)
        self.actions_grading["snapshot"].setEnabled(False)
        self.actions_grading["manual_detect"].setEnabled(False)
        self.actions_grading["edit_id"].setEnabled(True)
        self.actions_grading["continue"].setEnabled(True)
        self.actions_grading["discard"].setEnabled(True)
        self.actions_session["new"].setEnabled(False)
        self.actions_session["open"].setEnabled(False)
        self.actions_session["close"].setEnabled(True)
        self.actions_session["exit"].setEnabled(True)
        self.actions_tools["camera"].setEnabled(False)
        self.actions_tools["export_exam_config"].setEnabled(True)
        self.actions_tools["edit_scores"].setEnabled(True)
        self.actions_exams["search"].setEnabled(False)
        self.actions_exams["students"].setEnabled(True)
        self.actions_exams["export"].setEnabled(True)
        self.actions_exams["statistics"].setEnabled(True)

    def set_before_review_from_grading_mode(self):
        self.actions_grading["start"].setEnabled(False)
        self.actions_grading["stop"].setEnabled(False)
        self.actions_grading["back"].setEnabled(False)
        self.actions_grading["snapshot"].setEnabled(False)
        self.actions_grading["manual_detect"].setEnabled(False)
        self.actions_grading["edit_id"].setEnabled(False)
        self.actions_grading["continue"].setEnabled(False)
        self.actions_grading["discard"].setEnabled(False)
        self.actions_session["new"].setEnabled(False)
        self.actions_session["open"].setEnabled(False)
        self.actions_session["close"].setEnabled(False)
        self.actions_session["exit"].setEnabled(True)
        self.actions_tools["camera"].setEnabled(False)
        self.actions_tools["export_exam_config"].setEnabled(False)
        self.actions_tools["edit_scores"].setEnabled(False)
        self.actions_exams["search"].setEnabled(False)
        self.actions_exams["students"].setEnabled(False)
        self.actions_exams["export"].setEnabled(False)
        self.actions_exams["statistics"].setEnabled(False)

    def set_review_from_session_mode(self):
        self.actions_grading["start"].setEnabled(True)
        self.actions_grading["stop"].setEnabled(False)
        self.actions_grading["back"].setEnabled(True)
        self.actions_grading["snapshot"].setEnabled(False)
        self.actions_grading["manual_detect"].setEnabled(False)
        self.actions_grading["edit_id"].setEnabled(True)
        self.actions_grading["continue"].setEnabled(True)
        self.actions_grading["discard"].setEnabled(True)
        self.actions_session["new"].setEnabled(False)
        self.actions_session["open"].setEnabled(False)
        self.actions_session["close"].setEnabled(True)
        self.actions_session["exit"].setEnabled(True)
        self.actions_tools["camera"].setEnabled(True)
        self.actions_tools["export_exam_config"].setEnabled(True)
        self.actions_tools["edit_scores"].setEnabled(True)
        self.actions_exams["search"].setEnabled(False)
        self.actions_exams["students"].setEnabled(True)
        self.actions_exams["export"].setEnabled(True)
        self.actions_exams["statistics"].setEnabled(True)

    def set_session_mode(self):
        self.actions_grading["start"].setEnabled(True)
        self.actions_grading["stop"].setEnabled(False)
        self.actions_grading["back"].setEnabled(False)
        self.actions_grading["snapshot"].setEnabled(False)
        self.actions_grading["manual_detect"].setEnabled(False)
        self.actions_grading["edit_id"].setEnabled(False)
        self.actions_grading["continue"].setEnabled(False)
        self.actions_grading["discard"].setEnabled(False)
        self.actions_session["new"].setEnabled(False)
        self.actions_session["open"].setEnabled(False)
        self.actions_session["close"].setEnabled(True)
        self.actions_session["exit"].setEnabled(True)
        self.actions_tools["camera"].setEnabled(True)
        self.actions_tools["export_exam_config"].setEnabled(True)
        self.actions_tools["edit_scores"].setEnabled(True)
        self.actions_exams["search"].setEnabled(True)
        self.actions_exams["search"].setEnabled(False)
        self.actions_exams["students"].setEnabled(True)
        self.actions_exams["export"].setEnabled(True)
        self.actions_exams["statistics"].setEnabled(True)

    def set_manual_detect_mode(self):
        self.actions_grading["start"].setEnabled(False)
        self.actions_grading["stop"].setEnabled(True)
        self.actions_grading["back"].setEnabled(False)
        self.actions_grading["snapshot"].setEnabled(False)
        self.actions_grading["manual_detect"].setEnabled(True)
        self.actions_grading["edit_id"].setEnabled(False)
        self.actions_grading["continue"].setEnabled(False)
        self.actions_grading["discard"].setEnabled(True)
        self.actions_session["new"].setEnabled(False)
        self.actions_session["open"].setEnabled(False)
        self.actions_session["close"].setEnabled(True)
        self.actions_session["exit"].setEnabled(True)
        self.actions_tools["camera"].setEnabled(False)
        self.actions_tools["export_exam_config"].setEnabled(True)
        self.actions_tools["edit_scores"].setEnabled(False)
        self.actions_exams["search"].setEnabled(False)
        self.actions_exams["students"].setEnabled(False)
        self.actions_exams["export"].setEnabled(False)
        self.actions_exams["statistics"].setEnabled(False)

    def set_no_session_mode(self):
        for key in self.actions_grading:
            self.actions_grading[key].setEnabled(False)
        self.actions_session["new"].setEnabled(True)
        self.actions_session["open"].setEnabled(True)
        self.actions_session["close"].setEnabled(False)
        self.actions_session["exit"].setEnabled(True)
        self.actions_tools["camera"].setEnabled(True)
        self.actions_tools["export_exam_config"].setEnabled(False)
        self.actions_tools["edit_scores"].setEnabled(False)
        for key in self.actions_exams:
            self.actions_exams[key].setEnabled(False)

    def enable_manual_detect(self, enabled):
        """Enables or disables the manual detection mode.

        If `enable` is True, it is enabled. Otherwise, it is disabled.

        """
        self.actions_grading["manual_detect"].setEnabled(enabled)

    def register_listener(self, key, listener):
        actions = self._select_action_group(key[0])
        assert key[1] in actions
        actions[key[1]].triggered.connect(listener)

    def is_action_checked(self, key):
        """For checkabel actions, returns whether the action is checked.

        Action keys are tuples such as ('tools', 'lines').

        """
        actions = self._select_action_group(key[0])
        assert key[1] in actions
        assert actions[key[1]].isCheckable()
        return actions[key[1]].isChecked()

    def _select_action_group(self, key):
        if key == "session":
            return self.actions_session
        elif key == "grading":
            return self.actions_grading
        elif key == "exams":
            return self.actions_exams
        elif key == "tools":
            return self.actions_tools
        elif key == "help":
            return self.actions_help
        assert False, "Undefined action group key: {0}".format(key)

    def _add_action(self, action_name, icon_file, text, shortcuts, group, actions_list):
        action = self._create_action(action_name, icon_file, text, shortcuts)
        if action_name.startswith("+"):
            if action_name.startswith("++"):
                action_name = action_name[2:]
            else:
                action_name = action_name[1:]
        if not action.isSeparator():
            group[action_name] = action
        actions_list.append(action)

    def _create_action(self, action_name, icon_file, text, shortcuts):
        if action_name == "*separator*":
            action = QAction(self.window)
            action.setSeparator(True)
        else:
            if icon_file:
                action = QAction(
                    QIcon(utils.resource_path(icon_file)), text, self.window
                )
            else:
                action = QAction(text, self.window)
        if shortcuts:
            sequences = [QKeySequence(s) for s in shortcuts]
            action.setShortcuts(sequences)
        if action_name.startswith("+"):
            action.setCheckable(True)
            if action_name.startswith("++"):
                action.setChecked(True)
        return action

    def _populate_menubar(self, action_lists):
        self.menus["session"] = QMenu(_("&Session"), self.menubar)
        self.menus["grading"] = QMenu(_("&Grading"), self.menubar)
        self.menus["exams"] = QMenu(_("&Exams"), self.menubar)
        self.menus["tools"] = QMenu(_("&Tools"), self.menubar)
        self.menus["help"] = QMenu(_("&Help"), self.menubar)
        self.menubar.addMenu(self.menus["session"])
        self.menubar.addMenu(self.menus["grading"])
        self.menubar.addMenu(self.menus["exams"])
        self.menubar.addMenu(self.menus["tools"])
        self.menubar.addMenu(self.menus["help"])
        for action in action_lists["session"]:
            self.menus["session"].addAction(action)
        for action in action_lists["grading"]:
            self.menus["grading"].addAction(action)
        for action in action_lists["exams"]:
            self.menus["exams"].addAction(action)
        for action in action_lists["tools"]:
            self.menus["tools"].addAction(action)
        for action in action_lists["help"]:
            self.menus["help"].addAction(action)

    def _populate_toolbar(self, action_lists):
        for action in action_lists["grading"]:
            self.toolbar.addAction(action)
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.actions_exams["export"])
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.actions_session["new"])
        self.toolbar.addAction(self.actions_session["open"])
        self.toolbar.addAction(self.actions_session["close"])

    def _add_debug_actions(self):
        actions_list = []
        for key, icon, text, shortcuts in ActionsManager._actions_debug_data:
            self._add_action(
                key, icon, text, shortcuts, self.actions_tools, actions_list
            )
        menu = QMenu(_("&Debug options"), self.menus["tools"])
        for action in actions_list:
            menu.addAction(action)
        self.menus["tools"].addMenu(menu)

    def _add_experimental_actions(self):
        if ActionsManager._actions_experimental:
            actions_list = []
            for key, icon, text, shortcuts in ActionsManager._actions_experimental:
                self._add_action(
                    key, icon, text, shortcuts, self.actions_tools, actions_list
                )
            menu = QMenu(_("&Experimental"), self.menus["tools"])
            for action in actions_list:
                menu.addAction(action)
            self.menus["tools"].addMenu(menu)


class CenterView(QWidget):
    img_correct = '<img src="%s" height="22" width="22">' % utils.resource_path(
        "correct.svg"
    )
    img_incorrect = '<img src="%s" height="22" width="22">' % utils.resource_path(
        "incorrect.svg"
    )
    img_unanswered = '<img src="%s" height="22" width="22">' % utils.resource_path(
        "unanswered.svg"
    )

    def __init__(self, parent=None):
        super().__init__(parent)
        layout = QVBoxLayout()
        self.setLayout(layout)
        self.center = QStackedLayout()
        self.camview = widgets.CamView((640, 480), self, draw_logo=True)
        self.label_up = QLabel()
        self.label_down = QLabel()
        self.center.addWidget(self.camview)
        layout.addLayout(self.center)
        layout.addWidget(self.label_up)
        layout.addWidget(self.label_down)
        self.adjustSize()
        self.setFixedSize(self.sizeHint())

    def update_status(
        self,
        score: scoring.Score,
        model: Optional[str] = None,
        seq_num: Optional[int] = None,
        survey_mode: bool = False,
    ) -> None:
        parts = []
        if score is not None:
            if not survey_mode:
                parts.append(CenterView.img_correct)
                if score.correct is not None:
                    parts.append(str(score.correct) + "  ")
                else:
                    parts.append("---  ")
                parts.append(CenterView.img_incorrect)
                if score.incorrect is not None:
                    parts.append(str(score.incorrect) + "  ")
                else:
                    parts.append("---  ")
                parts.append(CenterView.img_unanswered)
                if score.blank is not None:
                    parts.append(str(score.blank) + "  ")
                else:
                    parts.append("---  ")
                if score.score is not None and score.max_score is not None:
                    parts.append(
                        _("Score: {0:.2f} / {1:.2f}").format(
                            score.score, score.max_score
                        )
                    )
                    parts.append("  ")
            else:
                parts.append(_("[Survey mode on]"))
                parts.append("  ")
        if model is not None:
            parts.append(_("Model:") + " " + model + "  ")
        if seq_num is not None:
            parts.append(_("Num.:") + " " + str(seq_num) + "  ")
        self.label_down.setText(
            ('<span style="white-space: pre">' + " ".join(parts) + "</span>")
        )

    def update_text_up(self, text: str) -> None:
        self.label_up.setText(text)

    def update_text_down(self, text: str) -> None:
        self.label_down.setText(text)

    def display_capture(self, ipl_image) -> None:
        """Displays a captured image in the window.

        The image is in the OpenCV IPL format.

        """
        self.camview.display_capture(ipl_image)

    def display_wait_image(self) -> None:
        """Displays the default image instead of a camera capture."""
        self.camview.display_wait_image()

    def register_listener(
        self, key: Tuple[str, str], listener: Callable[[int, int], None]
    ) -> None:
        """Registers listeners for the center view.

        Available listeners are:

        - ('camview', 'mouse_pressed'): mouse pressed in the camview
          area. The listener receives the coordinates (x, y) as a
          tuple.

        """
        if key[0] == "camview":
            if key[1] == "mouse_pressed":
                self.camview.register_mouse_pressed_listener(listener)
            else:
                assert False, "Undefined listener key: {0}".format(key)
        else:
            assert False, "Undefined listener key: {0}".format(key)

    def save_as_image(self, filename: str) -> bool:
        return self.grab().save(filename)


class OfflineCaptureSaver:
    def __init__(self, exam: exams.Exam, survey_mode: bool):
        self.view = CenterView()
        self.view.display_capture(exam.capture.image_drawn)
        self.view.update_status(
            exam.score, exam.decisions.model, exam.exam_id, survey_mode
        )

    def save(self, filename: str) -> None:
        self.view.grab().save(filename)


class ModalProgressDialog:
    def __init__(self, labelText, maximum, parent=None):
        self.progress_dialog = QProgressDialog(
            labelText, None, 0, maximum, parent=parent
        )
        self.progress_dialog.setWindowModality(Qt.WindowModality.WindowModal)
        self.value = 0

    def count_step(self):
        self.value += 1
        self.progress_dialog.setValue(self.value)


class MainWindow(QMainWindow):
    def __init__(self):
        super().__init__()
        policy = QSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        self.setSizePolicy(policy)
        self.setStatusBar(widgets.StatusBar(self))
        self.center_view = CenterView()
        self.exams_view = examsview.ThumbnailsView(self)
        #        self.center_layout = QStackedLayout()
        self.center_layout = QHBoxLayout()
        self.center_layout.addWidget(self.center_view)
        self.center_layout.addWidget(self.exams_view)
        center_container = QWidget(self)
        center_container.setLayout(self.center_layout)
        self.setCentralWidget(center_container)
        self.setWindowTitle("Eyegrade")
        self.setWindowIcon(QIcon(utils.resource_path("logo.svg")))
        self.adjustSize()
        #        self.setFixedSize(self.sizeHint())
        self.digit_key_listener = None
        self.exit_listener = False

    def keyPressEvent(self, event):
        if (
            self.digit_key_listener
            and event.key() >= Qt.Key.Key_0
            and event.key() <= Qt.Key.Key_9
        ):
            self.digit_key_listener(event.text())

    def register_listener(self, key, listener):
        if key[0] == "key_pressed":
            if key[1] == "digit":
                self.digit_key_listener = listener
            else:
                assert False, "Undefined listener key: {0}".format(key)
        elif key[0] == "exit":
            self.exit_listener = listener
        elif key[0] == "exam":
            if key[1] == "selected":
                self.exams_view.selection_changed.connect(listener)
        else:
            assert False, "Undefined listener key: {0}".format(key)

    def closeEvent(self, event):
        accept = True
        if self.exit_listener is not None:
            accept = self.exit_listener()
        if accept:
            event.accept()
        else:
            event.ignore()

    def clear_exams_view(self):
        self.exams_view.clear_exams()

    def update_status_bar(self, text):
        self.statusBar().set_message(text)


class Interface:
    def __init__(self, app, id_enabled, id_list_enabled, preferred_styles=None):
        self.app = app
        self.id_enabled = id_enabled
        self.id_list_enabled = id_list_enabled
        self.last_score = None
        self.last_model = None
        self.manual_detect_enabled = False
        self.window = MainWindow()
        self.actions_manager = ActionsManager(self.window)
        self.activate_no_session_mode()
        self.window.show()
        self.register_listener(("actions", "session", "exit"), self.window.close)
        self.register_listener(("actions", "help", "about"), self.show_about_dialog)
        self._configure_qt_style(preferred_styles)

    def run(self):
        return self.app.exec()

    def set_manual_detect_enabled(self, enabled):
        self.manual_detect_enabled = enabled
        self.actions_manager.enable_manual_detect(enabled)

    def activate_search_mode(self):
        self.actions_manager.set_search_mode()
        self.window.exams_view.block_keyboard(True)
        self.update_text("", "")
        self.update_status_bar(_("Grading - Scanning exam"))

    def activate_review_mode(self, from_grading):
        if from_grading:
            self.actions_manager.set_review_from_grading_mode()
            self.update_status_bar(_("Grading - Reviewing exam"))
        else:
            self.actions_manager.set_review_from_session_mode()
            self.update_status_bar(_("Session open - Reviewing exam"))

    def activate_before_review_from_grading_mode(self):
        self.actions_manager.set_before_review_from_grading_mode()

    def activate_manual_detect_mode(self):
        self.actions_manager.set_manual_detect_mode()
        self.update_text(_("Click on the outer corners of the answer tables"), "")
        self.update_status_bar(_("Grading - Manual detection mode"))

    def activate_session_mode(self):
        self.actions_manager.set_session_mode()
        self.window.exams_view.block_keyboard(False)
        self.display_wait_image()
        self.update_text("", "")
        self.update_status_bar(_("Session open"))

    def activate_no_session_mode(self):
        self.actions_manager.set_no_session_mode()
        self.display_wait_image()
        self.update_text("", "")
        self.update_status_bar(_("No session: open or create a session to start"))
        self.window.clear_exams_view()

    def add_exams(self, exams):
        self.window.exams_view.add_exams(exams)

    def add_exam(self, exam):
        self.window.exams_view.add_exam(exam)

    def update_exam(self, exam):
        self.window.exams_view.update_exam(exam)

    def remove_exam(self, exam):
        self.window.exams_view.remove_exam(exam)

    def selected_exam(self):
        return self.window.exams_view.selected_exam()

    def select_next_exam(self):
        return self.window.exams_view.select_next_exam()

    def clear_selected_exam(self):
        return self.window.exams_view.clear_selected_exam()

    def get_exams(self):
        return list(self.window.exams_view.exams)

    def enable_manual_detect(self, enabled):
        """Enables or disables the manual detection mode.

        If `enable` is True, it is enabled. Otherwise, it is disabled.

        """
        self.actions_manager.enable_manual_detect(enabled)

    def update_status(
        self,
        score: scoring.Score,
        model: Optional[str] = None,
        seq_num: Optional[int] = None,
        survey_mode: bool = False,
    ) -> None:
        self.window.center_view.update_status(
            score, model=model, seq_num=seq_num, survey_mode=survey_mode
        )

    def update_text_up(self, text: str) -> None:
        if text is None:
            text = ""
        self.window.center_view.update_text_up(text)

    def update_text_down(self, text: str) -> None:
        if text is None:
            text = ""
        self.window.center_view.update_text_down(text)

    def update_status_bar(self, text: str) -> None:
        self.window.update_status_bar(text)

    def update_text(self, text_up: str, text_down: str) -> None:
        self.window.center_view.update_text_up(text_up)
        self.window.center_view.update_text_down(text_down)

    def register_listeners(self, listeners: Dict[Tuple[str, ...], Callable]) -> None:
        """Registers a dictionary of listeners for the events of the gui.

        The listeners are specified as a dictionary with pairs
        event_key->listener. Keys are tuples of strings such as
        ('action', 'session', 'close').

        """
        for key, listener in listeners.items():
            self.register_listener(key, listener)

    def register_listener(self, key: Tuple[str, ...], listener: Callable) -> None:
        """Registers a single listener for the events of the gui.

        Keys are tuples of strings such as ('action', 'session',
        'close').

        """
        if key[0] == "actions":
            self.actions_manager.register_listener(key[1:], listener)
        elif key[0] == "center_view":
            self.window.center_view.register_listener(key[1:], listener)
        elif key[0] == "window":
            self.window.register_listener(key[1:], listener)
        else:
            assert False, "Unknown event key {0}".format(key)

    def is_action_checked(self, action_key):
        """For checkabel actions, returns whether the action is checked.

        Action keys are tuples such as ('tools', 'lines').

        """
        return self.actions_manager.is_action_checked(action_key)

    def register_timer(self, time_delta, callback):
        """Registers a callback function to be run after time_delta ms."""
        timer = QTimer(self.window)
        timer.setSingleShot(True)
        timer.timeout.connect(callback)
        timer.setInterval(time_delta)
        timer.start()

    def run_later(self, callback, delay=0):
        """Registers a callback for immediate enqueueing in the event loop."""
        self.register_timer(delay, callback)

    def display_capture(self, ipl_image):
        """Displays a captured image in the window.

        The image is in the OpenCV IPL format.

        """
        self.window.center_view.display_capture(ipl_image)

    def save_capture(self, filename: str) -> bool:
        """Saves the current capture and its annotations to the given file."""
        return self.window.center_view.save_as_image(filename)

    def display_wait_image(self):
        """Displays the default image instead of a camera capture."""
        self.window.center_view.display_wait_image()

    def dialog_new_session(self, config_filename=None):
        """Displays a new session dialog.

        An initial value for the path of the .eye file can be passed as
        `config_filename`.

        The data introduced by the user is returned as a dictionary with
        keys `directory`, `config` and `id_list`. `id_list` may be None.

        The return value is None if the user cancels the dialog.

        """
        dialog = wizards.WizardNewSession(self.window, config_filename=config_filename)
        result = dialog.exec()
        if result == QDialog.DialogCode.Accepted:
            return dialog.values()
        else:
            return None

    def dialog_student_id(self, ranked_students, student_listings):
        """Displays a dialog to change the student id.

        Returns a student object with the option selected by the user.
        The return value is None if the user cancels the dialog.

        """
        dialog = students.DialogStudentId(
            self.window, ranked_students, student_listings
        )
        return dialog.exec()

    def dialog_open_session(self):
        """Displays an open session dialog.

        The filename of the session file is returned or None.

        """
        filename, __ = QFileDialog.getOpenFileName(
            self.window,
            _("Select the session file"),
            "",
            FileNameFilters.session_db,
            None,
            QFileDialog.Option.DontUseNativeDialog,
        )
        return filename if filename else None

    def dialog_edit_scores(self, exam_config: exams.ExamConfig) -> bool:
        """Displays a dialog to edit question scores.

        Returns True if scores changed, or False otherwise.

        """
        dialog = scores.DialogEditScores(exam_config)
        return dialog.exec()

    def dialog_camera_selection(self, capture_context):
        """Displays a camera selection dialog.

        `capture_context` is the detection.ExamCaptureContext object
        to be used.

        """
        dialog = dialogs.DialogCameraSelection(capture_context, self.window)
        return dialog.exec()

    def dialog_export_grades(self, helper):
        """Displays the dialog for exporting grades.

        `helper` is a `eyegrade.export.GradesExportHelper object.

        If accepted by the user, it returns True, else False.

        """
        dialog = export.DialogExportGrades(self.window, helper)
        return dialog.exec()

    def dialog_show_statistics(self, stats: stats_core.Stats):
        """Displays the dialog for showing statistics."""
        dialog = show_stats.DialogShowStats(self.window, stats)
        dialog.exec()

    def dialog_students(self, student_listings):
        """Displays the student list."""
        dialog = students.DialogStudents(self.window, student_listings)
        return dialog.exec()

    def dialog_export_exam_config(self):
        """Displays the dialog for exporting the current exam configuration.

        If accepted by the user, it returns the filename.
        Returns None if cancelled.

        """
        save_dialog = QFileDialog(
            parent=self.window,
            caption=_("Save exam configration as..."),
            filter=_("Exam configuration (*.eye)"),
        )
        save_dialog.setOptions(QFileDialog.Option.DontUseNativeDialog)
        save_dialog.setDefaultSuffix("eye")
        save_dialog.setFileMode(QFileDialog.FileMode.AnyFile)
        save_dialog.setAcceptMode(QFileDialog.AcceptMode.AcceptSave)
        filename = None
        if save_dialog.exec():
            filename_list = save_dialog.selectedFiles()
            if len(filename_list) == 1:
                filename = filename_list[0]
        return filename

    def show_information(self, message, title="Information"):
        """Displays an dialog with an informative message.

        The method blocks until the user closes the dialog.

        """
        QMessageBox.information(self.window, title, message)

    def show_error(self, message, title="Error"):
        """Displays an error dialog with the given message.

        The method blocks until the user closes the dialog.

        """
        QMessageBox.critical(self.window, title, message)

    def show_warning(self, message, is_question=True):
        """Displays a warning dialog.

        Returns True if the the user accepts and False otherwise.

        """
        if not is_question:
            result = QMessageBox.warning(self.window, _("Warning"), message)
            if result == QMessageBox.StandardButton.Ok:
                return True
            else:
                return False
        else:
            result = QMessageBox.warning(
                self.window,
                _("Warning"),
                message,
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                QMessageBox.StandardButton.No,
            )
            if result == QMessageBox.StandardButton.Yes:
                return True
            else:
                return False

    def show_progress_dialog(
        self, label_text: str, max_steps: int
    ) -> ModalProgressDialog:
        """Shows a modal progress dialog from 0 to `max_steps`

        It returns a ModalProgressDialog object in order for the caller
        to be able to update progress in steps of 1 unit.

        """
        return ModalProgressDialog(label_text, max_steps, parent=self.window)

    def run_worker(self, task, callback):
        """Runs a task in another thread.

        The `task` must be an object that implements a `run()`
        method. Completion is notified to the given `callback` function.

        """
        worker = Worker(task)
        worker.finished.connect(callback)
        QThreadPool.globalInstance().start(worker)

    def show_about_dialog(self):
        dialog = dialogs.DialogAbout(self.window)
        dialog.exec()

    def _configure_qt_style(self, preferred_styles):
        if preferred_styles is not None:
            for style_key in preferred_styles:
                style = QStyleFactory.create(style_key)
                if style is not None:
                    self.app.setStyle(style)
                    break



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#

import math
import itertools
import statistics


# Data representation:
# - points: tuples (x, y)
# - Lines: tuples (rho, theta) as returned by the Hough transform
#          or two points p1 and p2 in the line.
#


# Functions on points
#
def distance(p1, p2):
    """Returns the distance from p1 to p2."""
    return math.sqrt(
        (p1[0] - p2[0]) * (p1[0] - p2[0]) + (p1[1] - p2[1]) * (p1[1] - p2[1])
    )


def diff_points(p1, p2):
    """Returns a tuple (p1.x - p2.x, p1.y - p2.y)."""
    return (p1[0] - p2[0], p1[1] - p2[1])


def add_points(p1, p2):
    """Returns a tuple (p1.x + p2.x, p1.y + p2.y)."""
    return (p1[0] + p2[0], p1[1] + p2[1])


def multiply_vector(vector, factor):
    """Multiplies a vector of two coordinates by a given factor."""
    return (factor * vector[0], factor * vector[1])


def round_point(point):
    """Rounds the coordinates of the point to the nearest integers."""
    return int(round(point[0])), int(round(point[1]))


def angles_perpendicular(angle1, angle2):
    """Returns True if angles are perpendicular or almost perpendicular.

    There is a margin of +-0.1 radians in which they are still
    considered perpendicular.

    """
    return (
        abs(angle2 - angle1 - math.pi / 2) < 0.1
        or abs(angle2 - angle1 + math.pi / 2) < 0.1
    )


def points_closer_to_horizontal(p1, p2):
    """Returns True if the line joining the two points is closer to
    the horizontal than to the vertical.
    """
    d = diff_points(p1, p2)
    return abs(d[0]) > abs(d[1])


# Functions on two-dimensional vectors
#
def scalar_product(vector1, vector2):
    """Returns the scalar product of two vectors."""
    return vector1[0] * vector2[0] + vector1[1] * vector2[1]


def module(vector):
    """Returns the module of a vector, given its two coordinates."""
    return math.sqrt(vector[0] * vector[0] + vector[1] * vector[1])


def angle_cosine(vector1, vector2):
    """Returns cos(angle) where angle is the angle between
    vector1 and vector2"""
    return scalar_product(vector1, vector2) / module(vector1) / module(vector2)


# Funtions on lines defined by two points
#
def slope(p1, p2):
    """Returns the slope of the line from p1 to p2."""
    return float(p2[1] - p1[1]) / (p2[0] - p1[0])


def slope_inv(p1, p2):
    """Returns the inverse of the slope of the line from p1 to p2."""
    return float(p2[0] - p1[0]) / (p2[1] - p1[1])


def closer_points(p1, p2, offset):
    """Returns a pair of points that are in the line that joins
    p1 and p2, but closer than them by offset."""
    dx = p2[0] - p1[0]
    dy = p2[1] - p1[1]
    k = float(offset) / math.sqrt(dx * dx + dy * dy)
    return (
        (int(p1[0] + dx * k), int(p1[1] + dy * k)),
        (int(p2[0] - dx * k), int(p2[1] - dy * k)),
    )


def closer_points_rel(p1, p2, offset_ratio, abs_offset=0):
    """Returns a pair of points that are in the line that joins
    p1 and p2, but closer than them by the given ratio.

    Example: offset_ratio == 0.9 implies the distance between the
    new points is 0.9 the distance between the original points.
    Points go then closer by abs_offset on each side.

    """
    dx = p2[0] - p1[0]
    dy = p2[1] - p1[1]
    xoff = abs_offset * float(dx) / (abs(dx) + abs(dy))
    yoff = abs_offset * float(dy) / (abs(dx) + abs(dy))
    k = (1.0 - float(offset_ratio)) / 2
    return (
        (int(p1[0] + dx * k + xoff), int(p1[1] + dy * k + yoff)),
        (int(p2[0] - dx * k - xoff), int(p2[1] - dy * k - yoff)),
    )


def walk_line(p0, p1):
    """Returns a generator of points in the line from p0 to p1 such
    that they are connected (in the way they would do to draw a
    line. Order of points is not guaranteed: points may go p0->p1
    or p1->p0. See walk_line_ordered if points must go from p0 to
    p1."""
    # Bresenham's algorithm as found in Wikipedia
    x0, y0 = p0
    x1, y1 = p1
    steep = abs(y1 - y0) > abs(x1 - x0)
    if steep:
        # swap values:
        x0, y0 = y0, x0
        x1, y1 = y1, x1
    if x0 > x1:
        # swap values:
        x0, x1 = x1, x0
        y0, y1 = y1, y0
    deltax = x1 - x0
    deltay = abs(y1 - y0)
    error = deltax / 2
    y = y0
    ystep = 1 if y0 < y1 else -1
    for x in range(x0, x1 + 1):
        yield (x, y) if not steep else (y, x)
        error = error - deltay
        if error < 0:
            y = y + ystep
            error = error + deltax


def walk_line_ordered(p0, p1):
    """Wrapper for walk_line that guarantees that points go from p0 to p1."""
    x0, y0 = p0
    x1, y1 = p1
    if abs(y1 - y0) > abs(x1 - x0):
        reverse = y0 > y1
    else:
        reverse = x0 > x1
    if not reverse:
        return walk_line(p0, p1)
    else:
        return reversed([p for p in walk_line(p0, p1)])


def interpolate_line(p0, p1, num_points):
    """Returns a list of num_points points in the line from p0 to p1.

    The list includes p0 and p1 and points are ordered from p0 to
    p1.

    """
    num_divisions = num_points - 1
    dx = float(p1[0] - p0[0]) / num_divisions
    dy = float(p1[1] - p0[1]) / num_divisions
    points = [p0]
    for i in range(1, num_divisions):
        points.append(round_point((p0[0] + dx * i, p0[1] + dy * i)))
    points.append(p1)
    return points


def interpolate_line_progressive(p0, p1, num_points, factor):
    """Returns a list of num_points points in the line from p0 to p1.

    The list includes p0 and p1 and points are ordered from p0 to
    p1.

    Distances are progressive according to factor: if factor > 1
    there is more space in between the first points; if factor == 1 there
    is exactly the same space between points; if factor < 1, there is
    more space between the last points.

    This function is twice as slow as the non progressive version.
    When factor is 1 the other version should be used instead.

    """
    diff = diff_points(p1, p0)
    n = num_points - 1
    h1 = 2.0 / n / (factor + 1)
    delta = h1 * (factor - 1) / (n - 1)
    positions = [
        h1 * (i - 1) + 0.5 * delta * (i * i - 3 * i + 2)
        for i in range(1, num_points + 1)
    ]
    points = [
        round_point(add_points(p0, multiply_vector(diff, pos))) for pos in positions
    ]
    # Just in case it wasn't rounded to the proper value
    points[-1] = p1
    return points


# Functions on lines represented as rho, theta
#
def intersection(hline, vline):
    """Returns the intersection point of a (nearly) horizontal line
    with a (nearly) vertical line. Results may be wrong in
    other cases."""
    rho1, theta1 = hline
    rho2, theta2 = vline
    y = (rho1 * math.cos(theta2) - rho2 * math.cos(theta1)) / math.sin(theta1 - theta2)
    x = (rho2 - y * math.sin(theta2)) / math.cos(theta2)
    return round_point((x, y))


def line_point(line, x=None, y=None):
    """Returns a point in the line with the given x or y coordinate.
    Either x or y must be None. Throws division by zero exception
    if x is set and the line is vertical or if y is set and the
    line is horizontal."""
    assert (x is None) ^ (y is None)
    rho, theta = line
    if y is None:
        y = int((rho - x * math.cos(theta)) / math.sin(theta))
    else:
        x = int((rho - y * math.sin(theta)) / math.cos(theta))
    return round_point((x, y))


def point_is_valid(point, dimensions):
    """Returns True if the coordinates of the point are within the bounds
    of the image, False otherwise. Parameter dimensions is a tuple
    (image_width, image_height)."""
    if (
        point[0] >= 0
        and point[1] >= 0
        and point[0] < dimensions[0]
        and point[1] < dimensions[1]
    ):
        return True
    else:
        return False


def project_point(point, from_line, to_line):
    """Assuming that both lines are parallel, traces a perpendicular
    line from from_line to to_line that contains the given point,
    and returns the intersection point of that line with from_line.
    In order to simplify computation, the function does not
    consider the angle of to_line, i.e. it assumes the same angle
    as from_line. The result is therefore only an approximation,
    valid when angles are close enough."""
    theta = from_line[1]
    d = to_line[0] - from_line[0]
    delta_x = d * math.cos(theta)
    delta_y = d * math.sin(theta)
    return (point[0] + delta_x, point[1] + delta_y)


def min_rho_difference(lines):
    """Being lines a list of lines (rho, theta) sorted according to rho,
    from low to high, returns the minimum difference between two
    consecutive lines. The size of the list must be at least two."""
    assert len(lines) >= 2
    min_diff = lines[1][0] - lines[0][0]
    for i in range(2, len(lines)):
        diff = lines[i][0] - lines[i - 1][0]
        if diff < min_diff:
            min_diff = diff
    return min_diff


def discard_spurious_lines(lines, expected):
    """Discards the discordant line(s).

    Discards the line or lines that minimizes the variance
    of the distances between the set of the other lines
    (of size expected).

    The hypothesis is that, when more lines than expected
    were detected, the variance of the distances bewteen
    lines will be minimized when the spurious lines are
    out of the set of selected lines.

    Be aware of the number of possible combinations before
    calling this function. Having just one or two extra lines
    should generally be ok.

    """
    best_variance = math.inf
    for combination in itertools.combinations(lines, expected):
        diffs = [b[0] - a[0] for a, b in zip(combination[:-1], combination[1:])]
        variance = statistics.variance(diffs)
        if variance < best_variance:
            best_combination = combination
            best_variance = variance
    return best_combination


# Functions on rectangles
#
def rect_center(plu, pru, pld, prd):
    """Returns the center of the rectangle with the given corners.
    Note: l/r are for left, right; u, d are for up/down."""
    return round_point((float(plu[0] + prd[0]) / 2, float(plu[1] + prd[1]) / 2))


# Functions on angles (in radians)
#
def distance_closest_axis(angle, axes_angles):
    """Returns the distance of angle to the closest angle in 'axes_angles'

    Axes are computed in pi modulus (that is, 0 equals pi).
    'axes_angles' is an iterable of angles. All parameters are in
    radians. Distance is returned as an angle in radians. +/- pi
    variations of the angles are also tested. The angle is
    normalized to the range [0, pi).

    """
    axes_expanded = []
    for a in axes_angles:
        a = a % math.pi
        axes_expanded.append(a)
        if a > math.pi / 2:
            axes_expanded.append(a - math.pi)
        else:
            axes_expanded.append(a + math.pi)
    angle_normalized = angle % math.pi
    return min([abs(angle_normalized - aa) for aa in axes_expanded])



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#
import argparse
import sys

# Local imports
from .. import utils
from .. import scoring
from . import latex
from . import parser


EyegradeException = utils.EyegradeException

EyegradeException.register_error(
    "correct_weight_none",
    "The option '--incorrect-weight' was set. It requires\n"
    "'--correct-weight' to be also set.",
    "'--correct-weight' not set.",
)


def read_cmd_options():
    arg_parser = argparse.ArgumentParser(
        description="Create exam PDF files using the LaTeX system."
    )
    arg_parser.add_argument("template", help="LaTeX template file")
    arg_parser.add_argument(
        "--version",
        action="version",
        version="{} {}".format(utils.program_name, utils.version),
    )
    arg_parser.add_argument(
        "-o",
        "--output-file-prefix",
        dest="output_file_prefix",
        help="store the output in the given file",
        default=None,
    )
    arg_parser.add_argument(
        "-e",
        "--exam",
        dest="exam_filename",
        default=None,
        help="filename of the questions for the exam",
    )
    arg_parser.add_argument(
        "-q",
        "--num-questions",
        type=int,
        dest="num_questions",
        help="number of questions",
        default=None,
    )
    arg_parser.add_argument(
        "-c",
        "--num-choices",
        type=int,
        dest="num_choices",
        help="number of choices per question",
        default=None,
    )
    arg_parser.add_argument(
        "-n",
        "--num-tables",
        type=int,
        dest="num_tables",
        help="number of answer tables",
        default=0,
    )
    arg_parser.add_argument("-d", "--date", dest="date", default=None, help="exam date")
    arg_parser.add_argument(
        "-s", "--subject", dest="subject", default=None, help="subject name"
    )
    arg_parser.add_argument(
        "-g", "--degree", dest="degree", default=None, help="degree name"
    )
    arg_parser.add_argument(
        "-m",
        "--models",
        dest="models",
        default="A",
        help="concatenation of the model leters to create",
    )
    arg_parser.add_argument(
        "-t", "--duration", dest="duration", default=None, help="exam duration time"
    )
    arg_parser.add_argument(
        "-l", "--title", dest="title", default=None, help="title of the exam"
    )
    arg_parser.add_argument(
        "-k",
        "--dont-shuffle-again",
        action="store_true",
        dest="dont_shuffle_again",
        default=False,
        help="don't shuffle already shuffled models",
    )
    arg_parser.add_argument(
        "--keep-question-order",
        action="store_true",
        dest="keep_question_order",
        default=False,
        help="keep the original order of the questions (shuffle choices only)",
    )
    arg_parser.add_argument(
        "-b",
        "--table-dimensions",
        dest="dimensions",
        default=None,
        help="table dimensions",
    )
    arg_parser.add_argument(
        "-f",
        "--force",
        dest="force_config_overwrite",
        action="store_true",
        default=False,
        help="force removal of the previous .eye exam file",
    )
    arg_parser.add_argument(
        "--cw",
        "--correct-weight",
        dest="correct_weight",
        help="score for correct answers",
        default=None,
    )
    arg_parser.add_argument(
        "--iw",
        "--incorrect-weight",
        dest="incorrect_weight",
        help="negative score for incorrect answers",
        default=None,
    )
    arg_parser.add_argument(
        "--id-length",
        type=int,
        dest="student_id_length",
        default=None,
        help="Number of digits of student IDs (0 to disable)",
    )
    arg_parser.add_argument(
        "--id-label",
        dest="student_id_label",
        default=None,
        help="Label to show with the student ID box",
    )
    # The -w below is maintained for compatibility; its use is deprecated
    # Use -W instead.
    arg_parser.add_argument(
        "-w",
        "-W",
        "--table-width",
        type=float,
        dest="table_width",
        default=None,
        help="answer table width in cm",
    )
    arg_parser.add_argument(
        "-H",
        "--table-height",
        type=float,
        dest="table_height",
        default=None,
        help="answer table height in cm",
    )
    arg_parser.add_argument(
        "-S",
        "--table-scale",
        type=float,
        dest="table_scale",
        default=1.0,
        help="scale answer table with respect to default"
        " values > 1.0 for augmenting, < 1.0 for reducing",
    )
    arg_parser.add_argument(
        "-x",
        "--id-box-width",
        type=float,
        dest="id_box_width",
        default=None,
        help="ID box width in cm",
    )
    arg_parser.add_argument(
        "--left-to-right-numbering",
        dest="left_to_right_numbering",
        action="store_true",
        default=False,
        help=("number questions from left to right instead of " "up to bottom"),
    )
    arg_parser.add_argument(
        "--survey-mode",
        dest="survey_mode",
        action="store_true",
        default=False,
        help=("this is a survey instead of an exam"),
    )
    arg_parser.add_argument(
        "--no-pdf",
        dest="no_pdf",
        action="store_true",
        default=False,
        help=("produce the .tex files instead of PDF"),
    )
    arg_parser.add_argument(
        "--variation",
        type=int,
        dest="variation",
        help="select the same variation for all questions (set 1 for the first variation)",
        default=None,
    )
    arg_parser.add_argument(
        "--keep-tex",
        dest="keep_tex",
        action="store_true",
        default=False,
        help="don't remove the .tex files after compilation",
    )
    args = arg_parser.parse_args()
    args.models = args.models.upper()
    # Either -e is specified, or -q and -c are used
    if not args.exam_filename:
        if args.dimensions is None and (
            args.num_questions is None or args.num_choices is None
        ):
            arg_parser.error(
                "A file with questions must be given, or options"
                " -q and -c must be set, or tables dimensions"
                " must be set."
            )
    else:
        if args.num_questions or args.num_choices:
            arg_parser.error("Option -e is mutually exclusive with -q and -c")
    # Check student id length 0 <= length <= 16
    if args.student_id_length is not None and (
        args.student_id_length < 0 or args.student_id_length > 16
    ):
        arg_parser.error(
            "The number of digits of student IDs must be "
            "between 0 and 16 (both included)"
        )
    # The scale factor must be greater than 0.1
    if args.table_scale < 0.1:
        arg_parser.error(
            "The scale factor must be positive and greater or equal" " to 0.1"
        )
    # Check score weights
    if args.correct_weight is not None:
        if args.incorrect_weight is None:
            args.incorrect_weight = 0
    elif args.incorrect_weight is not None:
        arg_parser.error("The score for correct answers is also needed (--cw)")
    return args


def create_exam():
    args = read_cmd_options()
    template_filename = args.template
    variables = {"subject": "", "degree": "", "date": "", "duration": "", "title": ""}
    scores = None
    dimensions = None

    # Take options from the input question files
    if args.exam_filename:
        exam = parser.parse_exam(args.exam_filename)
        if exam.subject is not None:
            variables["subject"] = exam.subject
        if exam.degree is not None:
            variables["degree"] = exam.degree
        if exam.date is not None:
            variables["date"] = exam.date
        if exam.duration is not None:
            variables["duration"] = exam.duration
        if exam.title is not None:
            variables["title"] = exam.title
        if exam.student_id_label is not None:
            variables["student_id_label"] = exam.student_id_label
        if exam.student_id_length is not None:
            variables["student_id_length"] = exam.student_id_length
        if exam.scores is not None:
            scores = exam.scores
        num_questions = exam.num_questions()
        num_choices = exam.num_choices()
        if not exam.homogeneous_num_choices():
            print(
                ("Warning: not all the questions have " "the same number of choices."),
                file=sys.stderr,
            )
        if num_choices is None:
            raise Exception(
                "All the questions in the exam must have the " "same number of choices"
            )
        for q in exam.questions:
            if q.num_correct_choices != 1:
                raise Exception("Questions must have exactly 1 correct choice")
    else:
        exam = None
        if args.dimensions is not None:
            dimensions, __ = utils.parse_dimensions(args.dimensions, True)
            num_choices = dimensions[0][0]
            num_questions = sum([d[1] for d in dimensions])
            if args.num_choices is not None and args.num_choices != num_choices:
                raise Exception("Incoherent number of choices")
            if args.num_questions is not None and args.num_questions != num_questions:
                raise Exception("Incoherent number of questions")
            if args.num_tables != 0 and len(dimensions) != args.num_tables:
                raise EyegradeException("", "incoherent_num_tables")
        else:
            num_questions = args.num_questions
            num_choices = args.num_choices
            if num_questions is None or num_choices is None:
                raise Exception("Expected a number of questions and choices")

    # Command line options override options from the file
    if args.date is not None:
        variables["date"] = args.date
    if args.subject is not None:
        variables["subject"] = args.subject
    if args.degree is not None:
        variables["degree"] = args.degree
    if args.title is not None:
        variables["title"] = args.title
    if args.duration is not None:
        variables["duration"] = args.duration
    if args.student_id_label is not None:
        variables["student_id_label"] = args.student_id_label
    if args.student_id_length is not None:
        variables["student_id_length"] = args.student_id_length
    if args.correct_weight is not None:
        scores = scoring.QuestionScores(args.correct_weight, args.incorrect_weight, 0)
    if args.output_file_prefix is None:
        output_file = sys.stdout
        config_filename = None
    else:
        config_filename = args.output_file_prefix + ".eye"
        if not args.survey_mode:
            output_file = args.output_file_prefix + "-%s.tex"
        else:
            output_file = args.output_file_prefix
    if args.variation is None:
        variation = None
    else:
        # Variations in cmd agrs begin at 1. Internally, they begin at 0.
        variation = args.variation - 1

    # Create and call the exam maker object
    maker = latex.ExamMaker(
        num_questions,
        num_choices,
        template_filename,
        output_file,
        variables,
        config_filename,
        num_tables=args.num_tables,
        dimensions=dimensions,
        table_width=args.table_width,
        table_height=args.table_height,
        table_scale=args.table_scale,
        id_box_width=args.id_box_width,
        force_config_overwrite=args.force_config_overwrite,
        scores=scores,
        points_words=exam.points_words,
        left_to_right_numbering=args.left_to_right_numbering,
        survey_mode=args.survey_mode,
    )
    if not args.no_pdf and args.output_file_prefix is not None:
        if latex.check_latex():
            produce_pdf = True
        else:
            produce_pdf = False
            print("Warning: pdflatex not found in your system PATH", file=sys.stderr)
    else:
        produce_pdf = False
    if exam is not None:
        maker.set_exam_questions(exam)
    if not args.survey_mode:
        for model in args.models:
            produced_filename = maker.create_exam(
                model,
                not args.dont_shuffle_again,
                variation=variation,
                produce_pdf=produce_pdf,
                keep_question_order=args.keep_question_order,
                keep_tex=args.keep_tex,
            )
            if produced_filename is not None:
                print("Created file:", produced_filename, file=sys.stderr)
        if args.output_file_prefix is not None:
            maker.output_file = args.output_file_prefix + "-%s-solutions.tex"
            for model in args.models:
                produced_filename = maker.create_exam(
                    model,
                    False,
                    with_solution=True,
                    variation=variation,
                    produce_pdf=produce_pdf,
                    keep_question_order=args.keep_question_order,
                    keep_tex=args.keep_tex,
                )
                print("Created file:", produced_filename, file=sys.stderr)
    else:
        produced_filename = maker.create_exam(
            None,
            not args.dont_shuffle_again,
            variation=variation,
            produce_pdf=produce_pdf,
            keep_question_order=args.keep_question_order,
            keep_tex=args.keep_tex,
        )
        print("Created file:", produced_filename, file=sys.stderr)
    if config_filename is not None:
        maker.save_exam_config()

    # Dump some final warnings
    for key in maker.empty_variables:
        print("Warning: empty '%s' variable" % key, file=sys.stderr)


def main():
    try:
        create_exam()
    except utils.EyegradeException as ex:
        print(ex, file=sys.stderr)


if __name__ == "__main__":
    main()



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#
import json
import argparse

from . import sample
from . import classifiers
from . import evaluation


def save_metadata(filename, metadata):
    with open(filename, mode="w") as f:
        json.dump(metadata, f, indent=4, sort_keys=True)


def k_fold_cross_evaluation(classifier, sample_set, rounds):
    classifier.reset()
    partitions = sample_set.partition(rounds)
    e = evaluation.KFoldCrossEvaluation(classifier, partitions)
    return e


def train_with_all(classifier, sample_set):
    classifier.reset()
    classifier.train(sample_set.samples())


def create_digit_classifier(sample_set, rounds):
    classifier = classifiers.DefaultDigitClassifier(
        load_from_file=None, confusion_matrix_from_file=None
    )
    e = k_fold_cross_evaluation(classifier, sample_set, rounds)
    metadata = {
        "performance": {
            "success_rate": e.success_rate,
            "balanced_success_rate": e.success_rate_balanced,
            "evaluation_rounds": rounds,
            "num_samples": len(sample_set),
        },
        "confusion_matrix": e.confusion_matrix_r.tolist(),
    }
    print(
        "Success rate: {} (balanced: {})".format(
            e.success_rate, e.success_rate_balanced
        )
    )
    save_metadata(classifiers.DEFAULT_DIG_META_FILE, metadata)
    train_with_all(classifier, sample_set)
    classifier.save(classifiers.DEFAULT_DIG_CLASS_FILE)


def create_crosses_classifier(sample_set, rounds):
    classifier = classifiers.DefaultCrossesClassifier(load_from_file=None)
    e = k_fold_cross_evaluation(classifier, sample_set, rounds)
    print(
        "Success rate: {} (balanced: {})".format(
            e.success_rate, e.success_rate_balanced
        )
    )
    metadata = {
        "performance": {
            "success_rate": e.success_rate,
            "balanced_success_rate": e.success_rate_balanced,
            "evaluation_rounds": rounds,
            "num_samples": len(sample_set),
        },
        "confusion_matrix": e.confusion_matrix_r.tolist(),
    }
    save_metadata(classifiers.DEFAULT_CROSS_META_FILE, metadata)
    train_with_all(classifier, sample_set)
    classifier.save(classifiers.DEFAULT_CROSS_CLASS_FILE)


def _parse_args():
    parser = argparse.ArgumentParser(
        description="Create the SVM classifier for digits or crosses."
    )
    parser.add_argument(
        "classifier", help='classifier to be created ("digits" or "crosses")'
    )
    parser.add_argument(
        "sample_files",
        metavar="sample file",
        nargs="+",
        help="index file with the samples for training/evaluation",
    )
    parser.add_argument(
        "--rounds",
        type=int,
        default=10,
        help="number of rounds for k-fold cross evaluation (default 100)",
    )
    return parser.parse_args()


def main():
    args = _parse_args()

    # Load the sample set:
    sample_set = sample.SampleSet()
    for filename in args.sample_files:
        sample_set.load_from_loader(sample.SampleLoader(filename))

    # Perform a k-fold cross-evaluation and create the classifier:
    if args.classifier == "digits":
        create_digit_classifier(sample_set, args.rounds)
    else:
        create_crosses_classifier(sample_set, args.rounds)


if __name__ == "__main__":
    main()



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#
import gettext

from PyQt6.QtWidgets import (
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QFileDialog,
    QFormLayout,
    QVBoxLayout,
    QWidget,
)

from .widgets import LabelledCheckBox
from . import FileNameFilters
from .. import utils
from .. import export


t = gettext.translation("eyegrade", utils.locale_dir(), fallback=True)
_ = t.gettext


class DialogExportGrades(QDialog):
    """Dialog to export grades listings.

    `helper` is a `eyegrade.export.GradesExportHlper object
    ready to be configured. When the dialog finishes it will
    contain all the options selected by the user.

    Example (replace `parent` by the parent widget):

    helper = eyegrade.export.GradesExportHelper(exam_config, student_groups)
    dialog = DialogExportGrades(parent, helper)
    result = dialog.exec()
    if result:
        # dialog accepted
    else:
        # dialog cancelled

    """

    def __init__(self, parent, helper):
        super().__init__(parent)
        self.helper = helper
        student_groups = helper.student_groups
        self.setWindowTitle(_("Export grades listing"))
        self.type_combo = QComboBox(parent=self)
        self.type_combo.addItem(_("Excel spreadsheet (.xlsx)"))
        self.type_combo.addItem(_("Tabs-separated file"))
        self.students_combo = QComboBox(parent=self)
        self.students_combo.addItem(_("All the students in the list"))
        self.students_combo.addItem(_("Only the students who attended the exam"))
        self.sort_combo = QComboBox(parent=self)
        self.sort_combo.addItem(_("Student list"))
        self.sort_combo.addItem(_("Last name"))
        self.sort_combo.addItem(_("Exam grading sequence"))
        if not student_groups:
            if not helper.survey_mode:
                self.students_combo.addItem(_("All the exams"))
            else:
                self.students_combo.addItem(_("All the surveys"))
            self.students_combo.setCurrentIndex(2)
            self.students_combo.setEnabled(False)
            self.sort_combo.setCurrentIndex(2)
            self.sort_combo.setEnabled(False)
        if len(student_groups) > 1:
            self.groups_combo = QComboBox(parent=self)
            self.groups_combo.addItem(_("All groups (one sheet)"))
            self.groups_combo.addItem(_("All groups (separate sheets)"))
            for group in student_groups:
                label = "{0} #{1.identifier} ({1.name})".format(_("Group"), group)
                self.groups_combo.addItem(label)
        else:
            self.groups_combo = None
        self.headers_checkbox = LabelledCheckBox(
            _("Add column headers"), self, checked=True
        )
        self.export_items = ExportItems(helper.survey_mode, len(student_groups), self)
        buttons = QDialogButtonBox(
            (
                QDialogButtonBox.StandardButton.Ok
                | QDialogButtonBox.StandardButton.Cancel
            )
        )
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)
        layout = QFormLayout(parent=self)
        self.setLayout(layout)
        layout.addRow(_("File type:"), self.type_combo)
        layout.addRow(_("Students:"), self.students_combo)
        if self.groups_combo is not None:
            layout.addRow(_("Student group:"), self.groups_combo)
        layout.addRow(_("Sort by:"), self.sort_combo)
        layout.addRow(_("Options:"), self.headers_checkbox)
        layout.addRow(_("Export fields:"), self.export_items)
        layout.addRow(buttons)

    def exec(self):
        """Shows the dialog and waits until it is closed.

        Returns True if the dialog gets accepted, or False otherwise.

        """
        result = False
        dialog_result = super().exec()
        if dialog_result == QDialog.DialogCode.Accepted:
            if self.type_combo.currentIndex() == 0:
                self.helper.file_format = export.FileFormat.XLSX
            elif self.type_combo.currentIndex() == 1:
                self.helper.file_format = export.FileFormat.CSV_TABS
            self.helper.file_name = self._get_save_file_name(self.helper.file_format)
            if self.helper.file_name:
                result = True
                if self.groups_combo is not None:
                    idx = self.groups_combo.currentIndex()
                    if idx > 1:
                        self.helper.export_group(idx - 2)
                    else:
                        self.helper.export_all_groups(idx == 0)
                elif self.helper.student_groups:
                    self.helper.export_group(0)
                else:
                    self.helper.export_all_exams()
                idx = self.sort_combo.currentIndex()
                if idx == 0:
                    self.helper.sort_by = export.SortBy.STUDENT_LIST
                elif idx == 1:
                    self.helper.sort_by = export.SortBy.LAST_NAME
                elif idx == 2:
                    self.helper.sort_by = export.SortBy.GRADING_SEQUENCE
                self.helper.all_students = self.students_combo.currentIndex() == 0
                self.helper.add_column_headers = self.headers_checkbox.is_checked()
                self.helper.export_columns(self.export_items.get_column_keys())
        return result

    def _get_save_file_name(self, file_format):
        if file_format == export.FileFormat.XLSX:
            file_filter = FileNameFilters.xlsx_file
            extension = "xlsx"
        else:
            file_filter = FileNameFilters.csv_file
            extension = "csv"
        save_dialog = QFileDialog(
            parent=self, caption=_("Save listing as..."), filter=file_filter
        )
        save_dialog.setOptions(QFileDialog.Option.DontUseNativeDialog)
        save_dialog.setDefaultSuffix(extension)
        save_dialog.setFileMode(QFileDialog.FileMode.AnyFile)
        save_dialog.setAcceptMode(QFileDialog.AcceptMode.AcceptSave)
        filename = None
        if save_dialog.exec():
            filename_list = save_dialog.selectedFiles()
            if len(filename_list) == 1:
                filename = filename_list[0]
        return filename


class ExportItems(QWidget):
    def __init__(self, survey_mode, there_are_students, parent):
        super().__init__(parent=parent)
        self.checkboxes = [
            (
                "student_id",
                LabelledCheckBox(
                    _("Student id number"),
                    self,
                    checked=not survey_mode,
                    enabled=there_are_students,
                ),
            ),
            (
                "name",
                LabelledCheckBox(
                    _("Student full name"),
                    self,
                    checked=not survey_mode,
                    enabled=there_are_students,
                ),
            ),
            (
                "last_name",
                LabelledCheckBox(
                    _("Student last name"),
                    self,
                    checked=False,
                    enabled=there_are_students,
                ),
            ),
            (
                "first_name",
                LabelledCheckBox(
                    _("Student first name"),
                    self,
                    checked=False,
                    enabled=there_are_students,
                ),
            ),
            (
                "exam_id",
                LabelledCheckBox(_("Exam sequence number"), self, checked=survey_mode),
            ),
            (
                "model",
                LabelledCheckBox(
                    _("Exam model letter"), self, checked=False, enabled=not survey_mode
                ),
            ),
            (
                "correct",
                LabelledCheckBox(
                    _("Number of correct answers"),
                    self,
                    checked=not survey_mode,
                    enabled=not survey_mode,
                ),
            ),
            (
                "incorrect",
                LabelledCheckBox(
                    _("Number of incorrect answers"),
                    self,
                    checked=not survey_mode,
                    enabled=not survey_mode,
                ),
            ),
            (
                "score",
                LabelledCheckBox(
                    _("Score"), self, checked=not survey_mode, enabled=not survey_mode
                ),
            ),
            (
                "answers",
                LabelledCheckBox(_("List of answers"), self, checked=survey_mode),
            ),
        ]
        layout = QVBoxLayout(self)
        for column, checkbox in self.checkboxes:
            layout.addWidget(checkbox)
        self.setLayout(layout)

    def get_column_keys(self):
        return [key for key, checkbox in self.checkboxes if checkbox.is_checked()]



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#
import argparse
import math

import numpy as np

from . import sample
from . import classifiers
from . import evaluation


def decide_params(classifier, sample_set, c_values, gamma_values, threshold=None, k=10):
    results = []
    rmat = np.zeros(shape=(len(c_values), len(gamma_values)), dtype="float32")
    partitions = sample_set.partition(k)
    for i, c in enumerate(c_values):
        for j, gamma in enumerate(gamma_values):
            params = dict(C=c, gamma=gamma)
            print("C: {}, gamma: {}".format(c, gamma))
            e = evaluation.KFoldCrossEvaluation(
                classifier, partitions, training_params=params, threshold=threshold
            )
            result = (e.success_rate, e.success_rate_balanced, c, gamma)
            results.append(result)
            rmat[i, j] = e.success_rate
            print(result)
            print(rmat)
    return results, rmat


def _parse_args():
    parser = argparse.ArgumentParser(description="Look for the best SVM parameters.")
    parser.add_argument(
        "classifier", help='classifier to be evaluated ("digits" or "crosses")'
    )
    parser.add_argument(
        "sample_files",
        metavar="sample file",
        nargs="+",
        help="index file with the samples for training/evaluation",
    )
    parser.add_argument(
        "--rounds",
        type=int,
        default=10,
        help="number of rounds for k-fold cross evaluation (default 10)",
    )
    return parser.parse_args()


def main():
    args = _parse_args()
    sample_set = sample.SampleSet()
    for filename in args.sample_files:
        sample_set.load_from_loader(sample.SampleLoader(filename))
    if args.classifier == "digits":
        classifier = classifiers.DefaultDigitClassifier(
            load_from_file=None, confusion_matrix_from_file=None
        )
        threshold = 0.9
    else:
        classifier = classifiers.DefaultCrossesClassifier(load_from_file=None)
        threshold = 0.99
    c_values = [math.pow(10, i) for i in np.linspace(0, 4, 9)]
    gamma_values = [math.pow(10, i) for i in np.linspace(-3, -1, 5)]
    r = decide_params(
        classifier, sample_set, c_values, gamma_values, threshold=threshold
    )
    print(r)


if __name__ == "__main__":
    main()



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#

import random

from typing import Dict, List, Set, Optional, Iterator, Tuple, Union, TYPE_CHECKING

from . import words
from .. import utils

if TYPE_CHECKING:
    from .. import scoring


utils.EyegradeException.register_error(
    "same_variation_selected_group",
    "The same variation must be selected for questions within the same group.",
)

utils.EyegradeException.register_error(
    "variation_index_out_of_range", "A variation index out of range was specified."
)


class ExamQuestions:
    questions: "QuestionsContainer"
    subject: Optional[str]
    degree: Optional[str]
    title: Optional[str]
    date: Optional[str]
    duration: Optional[str]
    student_id_label: Optional[str]
    shuffled_groups: Dict[str, List["QuestionsGroup"]]
    shuffled_questions: Dict[str, List["Question"]]
    permutations: Dict[str, List[int]]
    scores: Optional[Union["scoring.QuestionScores", "scoring.AutomaticScore"]]
    points_words: Optional[words.PointsWords]
    _student_id_length: int

    def __init__(self):
        self.questions = QuestionsContainer()
        self.subject = None
        self.degree = None
        self.title = None
        self.date = None
        self.duration = None
        self.student_id_label = None
        self._student_id_length = None
        self.shuffled_groups = {}
        self.shuffled_questions = {}
        self.permutations = {}
        self.scores = None
        self.points_words = None

    @property
    def student_id_length(self) -> int:
        return self._student_id_length

    @student_id_length.setter
    def student_id_length(self, length: int) -> None:
        if 0 <= length <= 16:
            self._student_id_length = length
        else:
            raise utils.EyegradeException(
                "Student id length must be bewteen " "0 and 16 (both included)"
            )

    def num_questions(self) -> int:
        """Returns the number of questions of the exam."""
        return len(self.questions)

    def num_choices(self) -> Optional[int]:
        """Returns the number of choices of the questions.

        If not all the questions have the same number of choices,
        it returns the maximum. If there are no exams, it returns None.

        """
        num = [q.num_choices for q in self.questions]
        if num:
            return max(num)
        return None

    def homogeneous_num_choices(self) -> Optional[bool]:
        """Returns True if all the questions have the same number of choices.

        Returns None if the list of questions is empty.

        """
        num = [q.num_choices for q in self.questions]
        if num:
            return min(num) == max(num)
        return None

    def shuffle(
        self, model: str, variation: Optional[int] = None, keep_question_order=False
    ) -> None:
        """Shuffles questions and options within questions for the given model."""
        if not keep_question_order:
            shuffled_groups, shuffled_questions, permutations = self.questions.shuffle()
        else:
            (
                shuffled_groups,
                shuffled_questions,
                permutations,
            ) = self.questions.dont_shuffle()
        self.shuffled_groups[model] = shuffled_groups
        self.shuffled_questions[model] = shuffled_questions
        self.permutations[model] = permutations
        if variation is None:
            self._shuffle_variations(model)
        else:
            self.select_variation(model, variation)
        for question in self.questions:
            question.shuffle(model)

    def _shuffle_variations(self, model: str) -> None:
        """Chooses a variation for each question randomly."""
        self.questions.shuffle_variations(model)

    def select_variations(self, model: str, variations: List[int]) -> None:
        if self.shuffled_groups is None:
            raise ValueError("Permutations muct be set before variations")
        for question, variation in zip(self.shuffled_questions[model], variations):
            question.select_variation(model, variation)
        # Select variations for groups also:
        i = 0
        for group in self.shuffled_groups[model]:
            group.select_variation(model, variations[i])
            i += len(group)
        self.questions.check_variations(model)

    def select_variation(self, model: str, variation: int) -> None:
        self.select_variations(model, [variation] * self.num_questions())

    def selected_variations(self, model: str) -> List[int]:
        return [
            question.selected_variation[model]
            for question in self.shuffled_questions[model]
        ]

    def selected_variation(self, model: str) -> Optional[int]:
        variations = []
        for question in self.shuffled_questions[model]:
            if len(question.variations) > 1:
                variations.append(question.selected_variation[model])
        if variations and min(variations) == max(variations):
            return variations[0]
        else:
            return None

    def set_permutation(
        self, model: str, permutations: List[Tuple[int, List[int]]]
    ) -> None:
        self.permutations[model] = [p[0] - 1 for p in permutations]
        self.shuffled_questions[model] = [
            self.questions[i] for i in self.permutations[model]
        ]
        self.shuffled_groups[model] = self._group_order(self.permutations[model])
        for question, permutation in zip(self.shuffled_questions[model], permutations):
            question.permutations[model] = [i - 1 for i in permutation[1]]

    def solutions_and_permutations(
        self, model: str
    ) -> Tuple[List[Set[int]], List[Tuple[int, List[int]]]]:
        solutions = []
        permutations = []
        for qid in self.permutations[model]:
            answers_perm = self.questions[qid].permutations[model]
            solutions.append({1 + answers_perm.index(0)})
            permutations.append((qid + 1, utils.increment_list(answers_perm)))
        return solutions, permutations

    def _group_order(self, permutations: List[int]) -> List["QuestionsGroup"]:
        """Computes group order from a permutations list."""
        group_for_question = {}
        i = 0
        for group in self.questions.groups:
            group_for_question[i] = group
            i += len(group)
        shuffled_groups = []
        for question in permutations:
            if question in group_for_question:
                shuffled_groups.append(group_for_question[question])
        return shuffled_groups


class QuestionsContainer:
    groups: List["QuestionsGroup"]

    def __init__(self):
        self.groups = []

    def __len__(self) -> int:
        return sum(len(group) for group in self.groups)

    def __iter__(self) -> Iterator["Question"]:
        return self._iterate_questions()

    def __getitem__(self, index: int) -> "Question":
        if index < 0:
            pos = len(self) - index
        else:
            pos = index
        i = 0
        for group in self.groups:
            if i + len(group) > pos:
                return group[pos - i]
            i += len(group)
        raise IndexError("QuestionsContainer index out of range: {}".format(index))

    def append(self, element: Union["QuestionsGroup", "Question"]) -> None:
        if isinstance(element, QuestionsGroup):
            self.groups.append(element)
        else:
            self.groups.append(QuestionsGroup([element]))

    def shuffle(self) -> Tuple[List["QuestionsGroup"], List["Question"], List[int]]:
        """Returns a tuple (list of questions, permutations) with data shuffled.

        Permutations is another list with the original position of each
        question. That is, question shuffled[i] was in the original list in
        permutations[i] position.

        """
        to_sort = [(random.random(), item) for item in self.groups]
        groups = []
        questions = []
        permutations = []
        for _, group in sorted(to_sort):
            groups.append(group)
            questions.extend(group.questions)
            permutations.extend(self._positions(group))
        return groups, questions, permutations

    def dont_shuffle(
        self,
    ) -> Tuple[List["QuestionsGroup"], List["Question"], List[int]]:
        """Returns a tuple (list of questions, permutations) with data not shuffled.

        Permutations is another list with the original position of each
        question. That is, question shuffled[i] was in the original list in
        permutations[i] position.

        """
        groups = []
        questions = []
        permutations = []
        for group in self.groups:
            groups.append(group)
            questions.extend(group.questions)
        permutations = list(range(len(questions)))
        return groups, questions, permutations

    def shuffle_variations(self, model: str) -> None:
        """Chooses a variation for each question randomly.

        Variations are chosen group by group:
        the same variation is chosen for all the questions of each group.

        """
        for group in self.groups:
            group.shuffle_variations(model)

    def check_variations(self, model: str) -> None:
        for group in self.groups:
            variations = [
                question.selected_variation[model] for question in group.questions
            ]
            if min(variations) < max(variations):
                raise utils.EyegradeException("", key="same_variation_selected_group")

    def _iterate_questions(self) -> Iterator["Question"]:
        for group in self.groups:
            for question in group:
                yield question

    def _positions(self, group: "QuestionsGroup") -> List[int]:
        pos = 0
        for other_group in self.groups:
            if other_group is group:
                return list(range(pos, pos + len(other_group)))
            pos += len(other_group)
        raise ValueError("Group not in group container")


class QuestionsGroup:
    questions: List["Question"]
    common_text: Optional["GroupCommonComponent"]

    def __init__(
        self,
        questions: List["Question"],
        common_text: Optional["GroupCommonComponent"] = None,
    ):
        if not questions:
            raise ValueError("Empty QuestionsGroup are not allowed")
        self.questions = questions
        self.common_text = common_text
        self._check_number_variations()

    @property
    def num_variations(self) -> int:
        # All questions have been checked to have
        # the same number of variations by the constructor.
        return self.questions[0].num_variations

    def __len__(self) -> int:
        return len(self.questions)

    def __iter__(self) -> Iterator["Question"]:
        return iter(self.questions)

    def __getitem__(self, index) -> "Question":
        return self.questions[index]

    def get_common_text(self, model: str) -> Optional["QuestionComponent"]:
        if self.common_text is not None:
            return self.common_text.component(model)
        return None

    def shuffle_variations(self, model: str) -> None:
        """Chooses a variation for this group randomly.

        All questions get the same variation.

        """
        self.select_variation(model, random.randrange(self.num_variations))

    def select_variation(self, model: str, variation: int) -> None:
        for question in self.questions:
            question.select_variation(model, variation)
        if self.common_text is not None:
            self.common_text.select_variation(model, variation)

    def __str__(self):
        return (
            f"<QuestionGroup: {len(self)} questions, {self.num_variations} variations>"
        )

    def _check_number_variations(self) -> None:
        num_variations_per_question = [
            question.num_variations for question in self.questions
        ]
        num_variations = min(num_variations_per_question)
        if num_variations < max(num_variations_per_question):
            raise utils.EyegradeException("", key="incompatible_variation_num_group")
        if self.common_text is not None:
            num_common_text_variations = self.common_text.num_variations
            if (
                num_common_text_variations > 1
                and num_common_text_variations != num_variations
            ):
                raise utils.EyegradeException(
                    "", key="incompatible_variation_num_group"
                )


class GroupCommonComponent:
    variations: List["QuestionComponent"]
    selected_variation: Dict[str, int]

    def __init__(self):
        self.variations = []
        self.selected_variation = {}

    @property
    def num_variations(self) -> int:
        return len(self.variations)

    def component(self, model: str) -> "QuestionComponent":
        return self.variations[self.selected_variation[model]]

    def add_variation(self, variation: "QuestionComponent") -> None:
        self.variations.append(variation)

    def select_variation(self, model: str, index: int) -> None:
        if index < 0:
            raise utils.EyegradeException(
                f"Variations: expected a value between 1 and {self.num_variations}, got {index + 1}",
                key="variation_index_out_of_range",
            )
        if index >= len(self.variations):
            if len(self.variations) == 1:
                index = 0
            else:
                raise utils.EyegradeException(
                    f"Variations: expected a value between 1 and {self.num_variations}, got {index + 1}",
                    key="variation_index_out_of_range",
                )
        self.selected_variation[model] = index


class FixedGroupCommonComponent(GroupCommonComponent):
    def __init__(self, common_component: "QuestionComponent"):
        super().__init__()
        self.add_variation(common_component)

    def component(self, model: str) -> "QuestionComponent":
        return self.variations[0]

    def add_variation(self, variation: "QuestionComponent") -> None:
        if self.variations:
            raise ValueError("Just one variation allowed in FixedGroupCommonComponent")
        super().add_variation(variation)


class Question:
    variations: List["QuestionVariation"]
    permutations: Dict[str, List[int]]
    selected_variation: Dict[str, int]

    def __init__(self):
        self.variations = []
        self.permutations = {}
        self.selected_variation = {}

    @property
    def num_choices(self) -> int:
        if not self.variations:
            raise ValueError("At least one question variation needed")
        return self.variations[0].num_choices

    @property
    def num_correct_choices(self) -> int:
        if not self.variations:
            raise ValueError("At least one question variation needed")
        return self.variations[0].num_correct_choices

    @property
    def tags(self) -> Optional[str]:
        if not self.variations:
            raise ValueError("At least one question variation needed")
        return self.variations[0].tags

    @property
    def num_variations(self) -> int:
        return len(self.variations)

    def text(self, model: str) -> "QuestionComponent":
        return self.variations[self.selected_variation[model]].text

    def shuffled_choices(self, model: str) -> List["QuestionComponent"]:
        if model == "0":
            return self.choices(model)
        return self.variations[self.selected_variation[model]].shuffled_choices(
            self.permutations[model]
        )

    def choices(self, model: str) -> List["QuestionComponent"]:
        return self.variations[self.selected_variation[model]].choices()

    def correct_choices(self, model: str) -> List["QuestionComponent"]:
        return self.variations[self.selected_variation[model]].correct_choices

    def add_variation(self, variation: "QuestionVariation") -> None:
        if self.variations and not self.variations[0].is_compatible(variation):
            raise utils.EyegradeException("", key="incompatible_variation")
        self.variations.append(variation)

    def shuffle(self, model: str) -> None:
        if not self.variations:
            raise ValueError("Cannot shuffle without at least one variation")
        self.permutations[model] = self.variations[0].shuffle()

    def select_variation(self, model: str, index: int) -> None:
        if index < 0:
            raise utils.EyegradeException(
                f"Variations: expected a value between 1 and {self.num_variations}, got {index + 1}",
                key="variation_index_out_of_range",
            )
        if index >= len(self.variations):
            if len(self.variations) == 1:
                index = 0
            else:
                raise utils.EyegradeException(
                    f"Variations: expected a value between 1 and {self.num_variations}, got {index + 1}",
                    key="variation_index_out_of_range",
                )
        self.selected_variation[model] = index


class FixedQuestion(Question):
    """A question without variations, i.e. just one variation."""

    def __init__(self, variation: "QuestionVariation"):
        super().__init__()
        self.add_variation(variation)

    @property
    def tags(self) -> Optional[str]:
        return self.variations[0].tags

    def text(self, model: str) -> "QuestionComponent":
        return self.variations[0].text

    def shuffled_choices(self, model: str) -> List["QuestionComponent"]:
        if model == "0":
            return self.choices(model)
        return self.variations[0].shuffled_choices(self.permutations[model])

    def choices(self, model: str) -> List["QuestionComponent"]:
        return self.variations[0].choices()

    def correct_choices(self, model: str) -> List["QuestionComponent"]:
        return self.variations[0].correct_choices

    def add_variation(self, variation: "QuestionVariation") -> None:
        if self.variations:
            raise ValueError("Just one variation allowed in FixedQuestion")
        super().add_variation(variation)


class QuestionVariation:
    text: "QuestionComponent"
    correct_choices: List["QuestionComponent"]
    incorrect_choices: List["QuestionComponent"]
    fix_first: List[int]
    fix_last: List[int]
    tags: Optional[str]

    def __init__(
        self,
        text: "QuestionComponent",
        correct_choices: List["QuestionComponent"],
        incorrect_choices: List["QuestionComponent"],
        fix_first: List[int],
        fix_last: List[int],
        tags: Optional[str] = None,
    ):
        self.text = text
        self.correct_choices = correct_choices
        self.incorrect_choices = incorrect_choices
        self.fix_first = fix_first
        self.fix_last = fix_last
        self.tags = tags

    @property
    def num_choices(self) -> int:
        return len(self.correct_choices) + len(self.incorrect_choices)

    @property
    def num_correct_choices(self) -> int:
        return len(self.correct_choices)

    def is_compatible(self, other: "QuestionVariation") -> bool:
        return (
            self.num_choices == other.num_choices
            and self.num_correct_choices == other.num_correct_choices
            and self.fix_first == other.fix_first
            and self.fix_last == other.fix_last
        )

    def shuffled_choices(self, permutation: List[int]) -> List["QuestionComponent"]:
        choices = self.correct_choices + self.incorrect_choices
        return [choices[i] for i in permutation]

    def choices(self) -> List["QuestionComponent"]:
        return self.correct_choices + self.incorrect_choices

    def shuffle(self) -> List[int]:
        to_sort = [(random.random(), pos) for pos in range(self.num_choices)]
        for pos in self.fix_first:
            # They will always be first when sorted
            to_sort[pos] = (-1.0, pos)
        for pos in self.fix_last:
            # They will always be last when sorted
            to_sort[pos] = (2.0, pos)
        permutations = []
        for _, pos in sorted(to_sort):
            permutations.append(pos)
        return permutations


class QuestionComponent:
    """A piece of text and optional figure or code.

    Represents both the text of a question and its choices.

    """

    in_choice: bool
    text: Optional[Union[str, List[Tuple[str, str]]]]
    code: Optional[str]
    figure: Optional[str]
    annex_width: Optional[float]
    annex_pos: Optional[str]
    annex_separation: bool

    def __init__(self, in_choice: bool, text: Optional[str] = None):
        self.in_choice = in_choice
        self.text = text
        self.code = None
        self.figure = None
        self.annex_width = None
        self.annex_pos = None

    def check_is_valid(self) -> None:
        if self.code is not None and self.figure is not None:
            raise Exception("Code and figure cannot be in the same block")
        if (
            self.in_choice
            and self.annex_pos != "center"
            and (self.code is not None or self.figure is not None)
        ):
            raise Exception("Figures and code in answers must be centered")
        if (
            self.code is not None
            and self.annex_pos == "center"
            and self.annex_width is not None
        ):
            raise Exception("Centered code cannot have width")
        if not self.in_choice and self.text is None:
            raise Exception("Questions must have a text")



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2024 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#

import pathlib

from .. import exams
from .. import sessiondb
from . import stats_core


def stats(exam_config: exams.ExamConfig, exams: list[exams.Exam]) -> stats_core.Stats:
    num_choices = max(exam_config.num_options)
    if exam_config.permutations:
        permutations = stats_core.ExamPermutations(
            exam_config.num_questions,
            num_choices,
            "0",
            exam_config.permutations,
        )
    else:
        permutations = None
    overall_stats = stats_core.OverallStats()
    q_by_q_stats = stats_core.QByQStats(
        exam_config.num_questions,
        num_choices,
        exam_config.solutions,
        permutations,
    )
    for exam in exams:
        overall_stats.count_answers(exam)
        q_by_q_stats.count_answers(exam.decisions.answers, exam.decisions.model)
    return stats_core.Stats(overall_stats.get_all_stats(), q_by_q_stats)


def stats_from_session_path(session_path: pathlib.Path) -> stats_core.Stats:
    with sessiondb.SessionDB(session_path, open=False) as session:
        return stats_from_session(session)


def stats_from_session(session: sessiondb.SessionDB) -> stats_core.Stats:
    return stats(session.exam_config, session.read_exams())



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Rodrigo Arguello, Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#

import cv2
import numpy as np
import numpy.linalg as linalg


class FeatureExtractor:
    """Default feature extractor.

    It assumes that images contain just one digit
    and ignores image corners.

    """

    def __init__(self, dim=28):
        self.dim = dim

    def extract(self, sample):
        image = self._reshape(sample)
        image = deskew(image, self.dim)
        image = clear_boundbox(image)
        image = cv2.resize(image, (self.dim, self.dim))
        image_matrix = np.array(image, np.float32) / 255.0
        feature_vector = image_matrix.reshape(self.features_len)
        return feature_vector

    @property
    def features_len(self):
        return self.dim * self.dim

    @staticmethod
    def _project_to_rectangle(sample, width, height):
        p = sample.corners
        corners_dst = np.array(
            [[0, 0], [width - 1, 0], [0, height - 1], [width - 1, height - 1]],
            dtype="float32",
        )
        h = cv2.findHomography(np.array(p, dtype="float32"), corners_dst)
        image = cv2.warpPerspective(sample.image, h[0], (width, height))
        return cv2.threshold(image, 64, 255, cv2.THRESH_BINARY)[1]

    @staticmethod
    def _reshape(sample):
        p = sample.corners
        width = int((cv2.norm(p[0, :], p[1, :]) + cv2.norm(p[2, :], p[3, :])) / 2)
        height = int((cv2.norm(p[0, :], p[2, :]) + cv2.norm(p[1, :], p[3, :])) / 2)
        return FeatureExtractor._project_to_rectangle(sample, width, height)


class CrossesFeatureExtractor(FeatureExtractor):
    """Feature extractor for crosses."""

    def __init__(self, dim=28):
        super().__init__(dim=dim)

    def extract(self, sample):
        image = self._project_to_rectangle(sample, self.dim, self.dim)
        #        image = cv2.resize(image, (self.dim, self.dim))
        image_matrix = np.array(image, np.float32) / 255.0
        feature_vector = image_matrix.reshape(self.features_len)
        return feature_vector


class OpenCVExampleExtractor:
    def __init__(self, dim=20, threshold=False):
        self.dim = dim
        self.threshold = threshold
        self._corners_dst = np.array(
            [[0, 0], [dim - 1, 0], [0, dim - 1], [dim - 1, dim - 1]], dtype="float32"
        )
        self.features_len = 64

    def extract(self, sample):
        corners = np.array(sample.corners, dtype="float32")
        h = cv2.findHomography(corners, self._corners_dst)
        image = cv2.warpPerspective(sample.image, h[0], (self.dim, self.dim))
        if self.threshold:
            image = cv2.threshold(image, 64, 255, cv2.THRESH_BINARY)[1]
        image = deskew(image, self.dim)
        feature_vector = self._preprocess_hog(image)
        return feature_vector

    def _preprocess_hog(self, image):
        gx = cv2.Sobel(image, cv2.CV_32F, 1, 0)
        gy = cv2.Sobel(image, cv2.CV_32F, 0, 1)
        mag, ang = cv2.cartToPolar(gx, gy)
        bin_n = 16
        bin_ = np.int32(bin_n * ang / (2 * np.pi))
        bin_cells = bin_[:10, :10], bin_[10:, :10], bin_[:10, 10:], bin_[10:, 10:]
        mag_cells = mag[:10, :10], mag[10:, :10], mag[:10, 10:], mag[10:, 10:]
        hists = [
            np.bincount(b.ravel(), m.ravel(), bin_n)
            for b, m in zip(bin_cells, mag_cells)
        ]
        hist = np.hstack(hists)
        # transform to Hellinger kernel
        eps = 1e-7
        hist /= hist.sum() + eps
        hist = np.sqrt(hist)
        hist /= linalg.norm(hist) + eps
        return np.float32(hist)


def deskew(image, dim):
    """Deskew an image.

    It improves classifier performance.
    The image must be a cv2 image.

    """
    affine_flags = cv2.WARP_INVERSE_MAP | cv2.INTER_LINEAR
    m = cv2.moments(image)
    if abs(m["mu02"]) < 1e-2:
        return image.copy()
    skew = m["mu11"] / m["mu02"]
    M = np.float32([[1, skew, -0.5 * dim * skew], [0, 1, 0]])
    image = cv2.warpAffine(image, M, (dim, dim), flags=affine_flags)
    return image


def clear_boundbox(image):
    """Clear the blank surrounding area of an image.

    It improves classifier performance.
    The image must be a cv2 image.

    """
    top = 0
    bot = image.shape[0]
    right = image.shape[1]
    left = 0
    it = 0
    for index, row in enumerate(image):
        if not np.all(row == 0) and it == 0:
            if index == image.shape[0] or not np.all(image[index + 1] == 0):
                top = index
                it = 1
        elif np.all(row == 0) and it == 1:
            bot = index
            break
    it = 0
    for index, col in enumerate(image.T):
        if (not np.all(col == 0)) and it == 0:
            if index + 2 >= image.shape[1] or not np.all(image.T[index + 2] == 0):
                left = index
                it = 1
        elif np.all(col == 0) and it == 1:
            right = index
            break
    cleared_image = image[top:bot, left:right]
    return cleared_image



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2018 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#
import sys
import random
import logging

import numpy as np
import cv2

from .. import sessiondb
from .. import detection
from .. import images
from ..ocr import sample

VALID_LABELS = (0, 1)


class LabeledCross:
    def __init__(self, label, image_file, corners, identifier=None):
        if not label in VALID_LABELS:
            raise ValueError("Wrong label value")
        self.label = label
        self.image_file = image_file
        self.corners = corners
        if identifier is not None:
            self.identifier = identifier
        else:
            self.identifier = random.randint(0, 100000000000000)

    def __str__(self):
        data = [self.image_file, str(self.label)]
        data.extend(str(n) for n in self.corners.reshape(8).tolist())
        return "\t".join(data)

    def crop(self):
        original = images.load_image(self.image_file)
        pre_processed = np.asarray(detection.pre_process(original)[:, :])
        samp = sample.CrossSampleFromCam(self.corners, pre_processed)
        cropped = samp.crop()
        cropped_file_path = "cross-{0}-{1}.png".format(self.label, self.identifier)
        cv2.imwrite(cropped_file_path, cropped.image)
        cropped_cross = LabeledCross(
            self.label, cropped_file_path, cropped.corners, identifier=self.identifier
        )
        return cropped_cross


def process_session(labeled_crosses, session_path):
    session = sessiondb.SessionDB(session_path)
    for exam in session.exams_iterator():
        image_file = session.get_raw_capture_path(exam["exam_id"])
        all_cells = session._read_answer_cells(exam["exam_id"])
        answers = session.read_answers(exam["exam_id"])
        for cells, answer in zip(all_cells, answers):
            crosses = [
                LabeledCross(
                    0, image_file, np.array([cell.plu, cell.pru, cell.pld, cell.prd])
                )
                for cell in cells
            ]
            if answer > 0:
                crosses[answer - 1].label = 1
            for cross in crosses:
                cropped_cross = cross.crop()
                if cropped_cross is not None:
                    labeled_crosses[cropped_cross.label].append(cropped_cross)
    session.close()


def dump_cross_list(labeled_crosses):
    with open("crosses.txt", "a") as f:
        for label in VALID_LABELS:
            for labeled_cross in labeled_crosses[label]:
                print(str(labeled_cross), file=f)


def _initialize_crosses_dict():
    labeled_crosses = {}
    for label in VALID_LABELS:
        labeled_crosses[label] = []
    return labeled_crosses


def main():
    logging.basicConfig(level=logging.INFO)
    labeled_crosses = _initialize_crosses_dict()
    for session_path in sys.argv[1:]:
        logging.info("Processing session {}".format(session_path))
        process_session(labeled_crosses, session_path)
    dump_cross_list(labeled_crosses)


if __name__ == "__main__":
    main()



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
import os
import unittest

import eyegrade.detection as detection


class _MockExamDetector(detection.ExamDetector):
    def __init__(self, dimensions):
        self.dimensions = dimensions
        self.status = {
            "lines": False,
            "boxes": False,
            "cells": False,
            "infobits": False,
            "id-box-hlines": False,
            "id-box": False,
        }
        self.success = False
        self.options = detection.ExamDetector.default_options
        self.image_to_show = None
        self.image_proc = None

    def _decide_cells(self, answer_cells):
        return None


def _mock_read_infobits(image, corner_matrixes):
    return [False, True, False, False, False, True]


detection.read_infobits = _mock_read_infobits


class TestDetection(unittest.TestCase):
    def _get_test_file_path(self, filename):
        dirname = os.path.dirname(os.path.abspath(__file__))
        return os.path.join(dirname, filename)

    def test_detect_capture(self):
        image_path = self._get_test_file_path("capture.png")
        options = detection.ExamDetector.get_default_options()
        options["capture-from-file"] = True
        options["capture-raw-file"] = image_path
        dimensions = ((3, 5),)
        for th in (170, 180, 190):
            context = detection.ExamDetectorContext(fixed_hough_threshold=th)
            detector = detection.ExamDetector(dimensions, context, options)
            success = detector.detect()
            if success:
                break
        self.assertTrue(success)
        self.assertEqual(len(detector.decisions.answers), 5)
        self.assertTrue(detector.decisions.answers[0] in range(4))
        self.assertTrue(detector.decisions.answers[1] in range(4))
        self.assertTrue(detector.decisions.answers[2] in range(4))
        self.assertTrue(detector.decisions.answers[3] in range(4))
        self.assertTrue(detector.decisions.answers[4] in range(4))
        self.assertEqual(len(detector.decisions.model), 1)

    def test_detect_capture_with_id(self):
        image_path = self._get_test_file_path("capture.png")
        options = detection.ExamDetector.get_default_options()
        options["capture-from-file"] = True
        options["capture-raw-file"] = image_path
        options["read-id"] = True
        options["id-num-digits"] = 9
        dimensions = ((3, 5),)
        for th in (170, 180, 190):
            context = detection.ExamDetectorContext(fixed_hough_threshold=th)
            detector = detection.ExamDetector(dimensions, context, options)
            success = detector.detect()
            if success:
                break
        self.assertTrue(success)
        self.assertEqual(len(detector.decisions.answers), 5)
        self.assertEqual(len(detector.decisions.detected_id), 9)
        self.assertEqual(len(detector.decisions.model), 1)

    def test_manual_detection(self):
        manual_points = [
            (113, 125),
            (251, 117),
            (304, 115),
            (447, 106),
            (117, 332),
            (259, 325),
            (312, 323),
            (458, 317),
        ]
        dimensions = [(3, 5), (3, 5)]
        detector = _MockExamDetector(dimensions)
        self.assertTrue(detector.detect_manual(manual_points))
        self.assertTrue(detector.success)
        corner_matrixes = detection.process_box_corners(manual_points, dimensions)
        # There are two answer boxes
        self.assertEqual(len(corner_matrixes), 2)
        # There are 5 + 1 = 6 lines in each answer box
        self.assertEqual(len(corner_matrixes[0]), 6)
        self.assertEqual(len(corner_matrixes[1]), 6)
        for box in corner_matrixes:
            for line in box:
                # There are 3 + 1 = 4 points in each line
                self.assertEqual(len(line), 4)
                for point in line:
                    self.assertEqual(len(point), 2)
        # Manual points are in the proper places:
        box_1, box_2 = corner_matrixes
        self.assertEqual(box_1[0][0], manual_points[0])
        self.assertEqual(box_1[0][3], manual_points[1])
        self.assertEqual(box_1[5][0], manual_points[4])
        self.assertEqual(box_1[5][3], manual_points[5])
        self.assertEqual(box_2[0][0], manual_points[2])
        self.assertEqual(box_2[0][3], manual_points[3])
        self.assertEqual(box_2[5][0], manual_points[6])
        self.assertEqual(box_2[5][3], manual_points[7])
        # Reordering points should have no effect:
        manual_points[1], manual_points[5] = manual_points[5], manual_points[1]
        manual_points[0], manual_points[4] = manual_points[4], manual_points[0]
        manual_points[3], manual_points[7] = manual_points[7], manual_points[3]
        self.assertTrue(detector.detect_manual(manual_points))
        corner_matrixes_2 = detection.process_box_corners(manual_points, dimensions)
        self.assertEqual(corner_matrixes, corner_matrixes_2)



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#

# The gettext module needs in Windows an environment variable
# to be defined before importing the gettext module itself
import os
import locale
import sys
import time
from typing import Optional
import webbrowser
import gettext

# Local imports
from . import detection
from . import utils
from . import exams
from .qtgui import gui
from . import sessiondb
from . import export
from .stats import compute_stats

if (
    not os.getenv("LANG")
    and not os.getenv("LANGUAGE")
    and not os.getenv("LC_MESSAGES")
    and not os.getenv("LC_ALL")
):
    lang, enc = locale.getdefaultlocale()
    if lang is not None:
        os.environ["LANG"] = lang

t = gettext.translation("eyegrade", utils.locale_dir(), fallback=True)
_ = t.gettext

utils.EyegradeException.register_error(
    "no_camera",
    _(
        "There is no suitable webcam. Eyegrade needs a webcam to work.\n"
        "If your computer already has a camera, check that it is not being\n"
        "used by another application."
    ),
)
utils.EyegradeException.register_error(
    "no_session_db", short_message=_("Session database file not found")
)
utils.EyegradeException.register_error(
    "session_db_exists",
    short_message=_("The session database cannot be created: " "it already exists"),
)
utils.EyegradeException.register_error(
    "session_invalid", short_message=_("The session database is not valid")
)
utils.EyegradeException.register_error(
    "corrupt_session_dir", short_message=_("The session directory has been corrupted.")
)
utils.EyegradeException.register_error(
    "incompatible_schema",
    short_message=_(
        "Incompatible session format. This is {0} version {1} "
        "but the session was created by version {2}"
    ),
)

param_fps = 16
capture_period = 1.0 / param_fps
capture_change_period = 0.3
capture_change_period_failure = 0.2
after_removal_delay = 1.0


class ImageDetectTask:
    """Used for running image detection in another thread."""

    def __init__(self, detector):
        self.detector = detector

    def run(self):
        self.detector.detect_safe()
        self.detector = None


class ImageChangeTask:
    """Used for running image change detection in another thread."""

    def __init__(self, detector, reference_image):
        self.detector = detector
        self.reference_image = reference_image

    def run(self):
        self.detector.try_to_detect()
        self.detector = None


class ManualDetectionManager:
    def __init__(self, exam, dimensions, detection_context, detector_options):
        self.exam = exam
        self.points = []
        self.detector = detection.ExamDetector(
            dimensions,
            detection_context,
            detector_options,
            image_raw=exam.capture.image_raw,
        )

    def add_point(self, point):
        self.points.append(point)

    def is_ready(self):
        return len(self.points) == 4 * len(self.detector.dimensions)

    def detect(self):
        assert self.is_ready()
        return self.detector.detect_manual(self.points)


class ProgramMode:
    """Represents the mode in which the program is."""

    no_session = 0
    session = 1
    search = 2
    review_from_session = 3
    review_from_grading = 4
    manual_detect = 5
    before_review_from_grading = 6

    def __init__(self):
        self.mode = ProgramMode.no_session

    def in_mode(self, mode):
        return self.mode == mode

    def in_no_session(self):
        return self.mode == ProgramMode.no_session

    def in_session(self):
        return self.mode == ProgramMode.session

    def in_search(self):
        return self.mode == ProgramMode.search

    def in_review(self):
        return (
            self.mode == ProgramMode.review_from_session
            or self.mode == ProgramMode.review_from_grading
        )

    def in_review_from_session(self):
        return self.mode == ProgramMode.review_from_session

    def in_review_from_grading(self):
        return self.mode == ProgramMode.review_from_grading

    def in_manual_detect(self):
        return self.mode == ProgramMode.manual_detect

    def in_before_review_from_grading(self):
        return self.mode == ProgramMode.before_review_from_grading

    def in_grading(self):
        return (
            self.in_search()
            or self.in_review_from_grading()
            or self.in_manual_detect()
            or self.in_before_review_from_grading()
        )

    def enter_mode(self, mode):
        self.mode = mode

    def enter_no_session(self):
        self.mode = ProgramMode.session

    def enter_session(self):
        self.mode = ProgramMode.no_session

    def enter_search(self):
        self.mode = ProgramMode.search

    def enter_review(self):
        if self.in_grading():
            self.mode = ProgramMode.review_from_grading
        else:
            self.mode = ProgramMode.review_from_session

    def enter_manual_detect(self):
        self.mode = ProgramMode.manual_detect

    def enter_before_review_from_grading(self):
        self.mode = ProgramMode.before_review_from_grading

    def __str__(self):
        if self.mode == ProgramMode.no_session:
            return "no_session"
        elif self.mode == ProgramMode.session:
            return "session"
        elif self.mode == ProgramMode.search:
            return "search"
        elif self.mode == ProgramMode.review_from_session:
            return "review_from_session"
        elif self.mode == ProgramMode.review_from_grading:
            return "review_from_grading"
        elif self.mode == ProgramMode.manual_detect:
            return "manual_detect"
        elif self.mode == ProgramMode.before_review_from_grading:
            return "before_review_from_grading"
        else:
            return f"unknown({self.mode})"


class ProgramManager:
    """Manages a grading session."""

    interface: gui.Interface
    session: Optional[sessiondb.SessionDB]
    exam: Optional[exams.Exam]

    def __init__(self, interface, session_file=None):
        self.interface = interface
        self.mode = ProgramMode()
        self.config = utils.config
        self.session = None
        self.detection_context = self._get_detection_context()
        self.detection_options = None
        self.drop_next_capture = False
        self.dump_buffer = False
        self._register_listeners()
        self.from_manual_detection = False
        self.manual_detect_manager = None
        self.exam = None
        if session_file is not None:
            self._try_session_file(session_file)

    def run(self):
        """Starts the program manager."""
        self.interface.run()

    def _get_detection_context(self):
        false_detector_session = os.getenv("EYEGRADE_CAMERA_SESSION")
        if not false_detector_session:
            return detection.ExamDetectorContext(camera_id=self.config["camera-dev"])
        else:
            return detection.FalseExamDetectorContext(false_detector_session)

    def _try_session_file(self, session_file):
        if os.path.isdir(session_file):
            filename = os.path.join(session_file, "session.eyedb")
            if os.path.exists(filename):
                session_file = filename
            else:
                self.interface.show_error(
                    _("The directory has no Eyegrade " "session") + ": " + session_file,
                    _("Error opening the session file"),
                )
                return
        valid, msg = self._validate_session(session_file)
        if valid:
            self._start_session()
        else:
            self.interface.show_error(msg)

    def _start_search_mode(self):
        self.mode.enter_search()
        self.from_manual_detection = False
        self.interface.activate_search_mode()
        self.exam = None
        self.latest_graded_exam = None
        self.latest_detector = None
        self.manual_detect_manager = None
        self.interface.register_timer(50, self._next_search)
        self.detection_context.dump_buffer(1.0)
        self.next_capture = time.monotonic() + 0.05

    def _start_review_mode(self):
        if self.mode.in_grading():
            self._store_exam(self.exam)
            # Automatic detection of exam removal to go to the next exam
            if self.interface.is_action_checked(("tools", "auto_change")):
                self._start_auto_change_detection()
            self.drop_next_capture = False
            # Because _store_capture_and_add is called with run_later (see below),
            # we have a temporary mode in which all actions are disabled.
            # This prevents, for example, trying to remove the exam before
            # it is saved and displayed.
            self.mode.enter_before_review_from_grading()
        else:
            self.mode.enter_review()
            self.interface.activate_review_mode(False)
        self.interface.display_capture(self.exam.get_image_drawn())
        self.interface.update_text_up(self.exam.get_student_id_and_name())
        if self.exam.score is not None:
            self.interface.update_status(
                self.exam.score,
                model=self.exam.decisions.model,
                seq_num=self.exam.exam_id,
                survey_mode=self.exam_data.survey_mode,
            )
        else:
            self.interface.update_text_down("")
        if not self.exam.decisions.model:
            self.interface.enable_manual_detect(True)
        if self.mode.in_grading():
            # Run later so that the widget gets fully painted before being
            # grabbed and saved.
            # The review_mode will be activated by _store_capture_and_add.
            self.interface.run_later(self._store_capture_and_add, delay=0)

    def _start_auto_change_detection(self):
        if not self.from_manual_detection:
            self.change_failures = 0
            self.interface.register_timer(1000, self._next_change_detection)

    def _start_manual_detect_mode(self):
        self.mode.enter_manual_detect()
        self.interface.activate_manual_detect_mode()
        self.interface.display_capture(self.exam.get_image_drawn())
        self.manual_detect_manager = ManualDetectionManager(
            self.exam,
            self.exam_data.dimensions,
            self.detection_context,
            self.detection_options,
        )

    def _next_search(self):
        if not self.mode.in_search():
            return
        if self.dump_buffer:
            self.dump_buffer = False
            self.detection_context.dump_buffer(after_removal_delay)
        detector = detection.ExamDetector(
            self.exam_data.dimensions, self.detection_context, self.detection_options
        )
        self.current_detector = detector
        task = ImageDetectTask(detector)
        self.interface.run_worker(task, self._after_image_detection)

    def _after_image_detection(self):
        detector = self.current_detector
        self.current_detector = None
        if not self.mode.in_search():
            # The user switched to other mode while the image was processed
            return
        self.latest_detector = detector
        if detector.status["boxes"] and self.detection_context.threshold_locked:
            self.detection_context.unlock_threshold()
        exam = self._process_capture(detector)
        if exam is None or not detector.success:
            if exam is not None:
                exam.draw_answers()
                exam.draw_status()
                self.interface.enable_manual_detect(False)
            elif detector.capture is not None:
                detector.capture.draw_status()
            if detector.capture is not None:
                self.interface.display_capture(detector.capture.image_drawn)
            self._schedule_next_capture(capture_period, self._next_search)
        elif not self.drop_next_capture:
            exam.draw_answers()
            self.exam = exam
            self._start_review_mode()
        else:
            # Special mode: do not lock until another capture is
            # available.  Used after auto exam removal detection.
            exam.draw_answers()
            self.interface.display_capture(detector.capture.image_drawn)
            self._schedule_next_capture(after_removal_delay, self._next_search)
            self.drop_next_capture = False
            self.dump_buffer = True

    def _next_change_detection(self):
        """Used to detect exam removal.

        This method captures an image and launches its analysis.
        Continuation of work is done at `_after_change_detection`.

        """
        if (
            not self.mode.in_review_from_grading()
            or not self.interface.is_action_checked(("tools", "auto_change"))
        ):
            return
        self.detection_context.dump_buffer(1.0)
        detector = detection.ExamDetector(
            self.exam_data.dimensions, self.detection_context, self.detection_options
        )
        self.current_detector = detector
        task = ImageChangeTask(detector, self.exam.capture)
        self.interface.run_worker(task, self._after_change_detection)

    def _after_change_detection(self):
        """Continuation of `_next_change_detection`.

        Executed after the image has been processed. This method decides
        whether the exam has been removed.

        """
        detector = self.current_detector
        self.current_detector = None
        if detector is None:
            # This needs to be investigated: this case should never
            # happen, but I saw it happen...
            return
        if (
            not self.mode.in_review_from_grading()
            or not self.interface.is_action_checked(("tools", "auto_change"))
        ):
            return
        exam_removed = False
        if detector.exam_detected:
            period = capture_change_period
            self.change_failures = 0
        else:
            period = capture_change_period_failure
            self.change_failures += 1
            if self.change_failures >= 4:
                exam_removed = True
        if not exam_removed:
            self._schedule_next_capture(period, self._next_change_detection)
        else:
            self.detection_context.lock_threshold()
            self.drop_next_capture = True
            self._action_continue()

    def _schedule_next_capture(self, period, function):
        """Schedules the next image capture and registers the timer.

        Call it from search mode, after an image has been processed, or
        in review mode if automatic exam removal detection is active.

        """
        current_time = time.monotonic()
        self.next_capture += period
        if current_time > self.next_capture:
            self.detection_context.dump_buffer((current_time - self.next_capture))
            wait = 0.010
            self.next_capture = time.monotonic() + 0.010
        else:
            wait = self.next_capture - current_time
        self.interface.register_timer(int(wait * 1000), function)

    def _process_capture(self, detector):
        """Processes a captured image."""
        exam = None
        if detector.status["infobits"] or not self.detection_options["infobits"]:
            if not self.detection_options["infobits"]:
                # If models are configured not to be detected, assume the default model "A"
                detector.decisions.model = "A"
            model = detector.decisions.model
            if model is not None:
                if model in self.exam_data.solutions or self.exam_data.survey_mode:
                    if model in self.exam_data.scores:
                        scores = self.exam_data.scores[model]
                    else:
                        scores = None
                    exam = exams.Exam(
                        detector.capture,
                        detector.decisions,
                        self.exam_data.get_solutions(model),
                        self.session.student_listings,
                        self.exam_id,
                        scores,
                        sessiondb=self.session,
                    )
                    self.latest_graded_exam = exam
                elif model not in self.exam_data.solutions:
                    msg = _("There are no solutions for model {0}.").format(model)
                    self.interface.show_error(msg)
        return exam

    def _new_session(self):
        """Callback for when the new session action is selected."""
        values = self.interface.dialog_new_session()
        if values is not None:
            self.exam_data = values["config"]
            self.exam_data.capture_pattern = self.config["save-filename-pattern"]
            try:
                sessiondb.create_session_directory(
                    values["directory"], self.exam_data, values["student_listings"]
                )
                self.session = sessiondb.SessionDB(values["directory"])
                self.session.capture_save_func = self.interface.save_capture
            except IOError as e:
                self.interface.show_error(_("Input/output error:") + " " + str(e))
            except utils.EyegradeException as e:
                self.interface.show_error(_("Error:") + " " + str(e))
            else:
                self._start_session()

    def _open_session(self):
        """Callback for when the open session action is selected."""
        filename = self.interface.dialog_open_session()
        if not filename:
            return
        valid, message = self._validate_session(filename)
        if not valid:
            self.interface.show_error(message)
        else:
            self._start_session()

    def _validate_session(self, filename):
        """Checks that the given session directory is valid.

        It loads and checks the exam configuration file from
        `filename` and checks the directories that should be at the
        directory. The file is stored as self.exam_data.

        Returns (valid, error_message) where exam_data is the exam
        configuration object, valid is True if the session is valid
        and error message contains a textal description of the error
        in case the session is not valid.

        """
        try:
            self.session = sessiondb.SessionDB(filename)
            self.exam_data = self.session.exam_config
            self.session.capture_save_func = self.interface.save_capture
            success = True
            message = ""
        except utils.EyegradeException as e:
            self.session = None
            self.exam_data = None
            success = False
            message = _("Error loading the session") + ": " + str(e)
        except IOError as e:
            self.session = None
            self.exam_data = None
            success = False
            message = _("Error loading the session") + ": " + str(e)
        return success, message

    def _close_session(self):
        """Callback that closes the current session."""
        if self.mode.in_manual_detect():
            if not self.interface.show_warning(
                _(
                    "The current capture has not been saved and will be lost. "
                    "Are you sure you want to close this session?"
                ),
                is_question=True,
            ):
                return
        if self.mode.in_grading() or self.mode.in_review_from_session():
            self._stop_grading()
        self.session.save_legacy_answers()
        self.session.close()
        self.mode.enter_no_session()
        self.session = None
        self.exam_data = None
        self.detection_options = None
        self.interface.activate_no_session_mode()

    def _action_edit_scores(self):
        """Callback for opening the question scores dialog"""
        current_exam = self.exam
        if self.mode.in_grading() or self.mode.in_review_from_session():
            self._stop_grading()
        self.interface.activate_session_mode()
        modified = self.interface.dialog_edit_scores(self.exam_data)
        if modified:
            exams = self.interface.get_exams()
            if exams:
                progress = self.interface.show_progress_dialog(
                    _("Updating the scores of already graded exams"), len(exams)
                )
                for exam in exams:
                    if exam.decisions.model in self.exam_data.scores:
                        question_scores = self.exam_data.scores[exam.decisions.model]
                    else:
                        question_scores = None
                    exam.update_question_scores(question_scores)
                    exam.load_capture()
                    exam.reset_image()
                    exam.draw_answers()
                    self.session.update_score(exam, commit=False)
                    self.session.save_drawn_capture(
                        exam.exam_id,
                        exam.capture,
                        exam.decisions.student,
                        gui.OfflineCaptureSaver(exam, self.exam_data.survey_mode),
                    )
                    self.interface.update_exam(exam)
                    progress.count_step()
                self.interface.show_information(
                    _("The scores of the already graded exams have been updated."),
                    title=_("Scores updated"),
                )
            self.session.update_exam_config_scores(self.exam_data, commit=True)
        if current_exam is not None:
            self._exam_selected(current_exam)

    def _exit_application(self):
        """Callback for when the user wants to exit the application."""
        if self.mode.in_manual_detect():
            exit_ = self.interface.show_warning(
                _(
                    "The current capture has not been saved and will be lost. "
                    "Are you sure you want to exit the application?"
                ),
                is_question=True,
            )
        elif self.mode.in_grading() or self.mode.in_review_from_session():
            self._stop_grading()
            exit_ = True
        else:
            exit_ = True
        if exit_ and self.session is not None:
            self.session.save_legacy_answers()
            self.session.close()
        return exit_

    def _action_start(self):
        if self.mode.in_review_from_session():
            self._stop_grading()
        self._start_grading()

    def _action_stop(self):
        self._stop_grading()

    def _action_back(self):
        if self.mode.in_review_from_session():
            self._stop_grading()
            self.interface.clear_selected_exam()
            self.exam = None

    def _action_snapshot(self):
        """Callback for the snapshot action."""
        if self.latest_graded_exam is None:
            if self.latest_detector is None:
                return
            detector = self.latest_detector
            self.exam = exams.Exam(
                detector.capture,
                detector.decisions,
                [],
                self.session.student_listings,
                self.exam_id,
                None,
                sessiondb=self.session,
            )
            self.exam.reset_image()
            enable_manual_detection = True
        else:
            self.exam = self.latest_graded_exam
            self.exam.reset_image()
            self.exam.draw_answers()
            enable_manual_detection = False
        self._start_review_mode()
        self.interface.enable_manual_detect(enable_manual_detection)

    def _action_discard(self):
        """Callback for cancelling/removing the current capture."""
        if self.mode.in_review_from_grading():
            self.session.remove_exam(self.exam.exam_id)
            self.interface.remove_exam(self.exam)
            self._start_search_mode()
        elif self.mode.in_manual_detect():
            self._start_search_mode()
        elif self.mode.in_review_from_session():
            remove = self.interface.show_warning(
                _("The selected exam will be removed. Are you sure?"), is_question=True
            )
            if remove:
                self.session.remove_exam(self.exam.exam_id)
                self.interface.remove_exam(self.exam)
                exam = self.interface.selected_exam()
                if exam is not None:
                    self._exam_selected(exam)
                else:
                    self.exam = None
                    self._activate_session_mode()

    def _action_continue(self):
        """Callback for saving the current capture."""
        if self.mode.in_review_from_grading():
            self.exam_id += 1
            self._start_search_mode()
        elif self.mode.in_review_from_session():
            self.interface.select_next_exam()

    def _action_manual_detect(self):
        """Callback for the manual detection action."""
        if self.mode.in_search():
            # Take the current snapshot and go to manual detect mode
            self.exam = exams.Exam(
                self.latest_detector.capture,
                self.latest_detector.decisions,
                [],
                self.session.student_listings,
                self.exam_id,
                None,
                sessiondb=self.session,
            )
            # Store the exam in order to emulate entering this mode
            # from review mode.
            self._store_exam(self.exam)
            self.interface.run_later(self._store_capture_and_add, delay=100)
        self.exam.reset_image()
        self._start_manual_detect_mode()

    def _action_edit_id(self):
        """Callback for the edit student id action."""
        if not self.mode.in_review():
            return
        students = self.exam.ranked_student_ids()
        student = self.interface.dialog_student_id(
            students, self.session.student_listings
        )
        if student is not None:
            self.exam.update_student_id(student)
            self.interface.update_text_up(self.exam.get_student_id_and_name())
            self.session.update_student(
                self.exam.exam_id,
                self.exam.capture,
                self.exam.decisions,
                store_captures=False,
            )
            self.interface.run_later(self._store_capture_and_update, delay=100)

    def _action_camera_selection(self):
        """Callback for opening the camera selection dialog."""
        self.interface.dialog_camera_selection(self.detection_context)

    def _action_help(self):
        """Callback for the help action."""
        webbrowser.open(utils.help_location, new=2)

    def _action_website(self):
        """Callback for the website action."""
        webbrowser.open(utils.web_location, new=2)

    def _action_source_code(self):
        """Callback for the website action."""
        webbrowser.open(utils.source_location, new=2)

    def _action_debug_changed(self):
        """Callback for the checkable actions in the debug options menu."""
        if self.detection_options is not None:
            self.detection_options["show-lines"] = self.interface.is_action_checked(
                ("tools", "lines")
            )
            self.detection_options["show-image-proc"] = (
                self.interface.is_action_checked(("tools", "processed"))
            )
            self.detection_options["show-status"] = self.interface.is_action_checked(
                ("tools", "show_status")
            )

    def _action_auto_change_changed(self):
        """Callback for the checkable 'auto_change' option."""
        if self.interface.is_action_checked(("tools", "auto_change")):
            self._start_auto_change_detection()

    def _action_export_grades(self):
        """Action for exporting the list of grades."""
        helper = export.GradesExportHelper(
            self.exam_data, self.session.get_student_groups()
        )
        result = self.interface.dialog_export_grades(helper)
        if result:
            try:
                self.session.export_grades(helper)
            except IOError as e:
                msg = _("Input/output error: {0}").format(e.strerror)
                self.interface.show_error(msg)
            else:
                self.interface.show_information(
                    _("The file has been saved."), title=_("File saved")
                )

    def _action_show_statistics(self):
        """Action for displaying a dialog with question by question statistics."""
        self.interface.dialog_show_statistics(
            compute_stats.stats_from_session(self.session)
        )

    def _action_students(self):
        student_listings = self.session.student_listings
        self.interface.dialog_students(student_listings)

    def _action_export_exam_config(self):
        """Callback for exporting the current exam configuration."""
        filename = self.interface.dialog_export_exam_config()
        if filename is not None:
            try:
                self.exam_data.save(filename)
            except IOError as e:
                msg = _("Input/output error: {0}").format(e.strerror)
                self.interface.show_error(msg)
            else:
                self.interface.show_information(
                    _("The file has been saved."), title=_("File saved")
                )

    def _mouse_pressed(self, point):
        """Callback called when the mouse is pressed inside a capture."""
        if self.mode.in_review():
            self._mouse_pressed_change_answer(point)
        elif self.mode.in_manual_detect():
            self._mouse_pressed_manual_detection(point)

    def _mouse_pressed_change_answer(self, point):
        if self.exam.capture.has_answer_cells():
            question, answer = self.exam.capture.get_cell_clicked(point)
            if question is not None:
                self.exam.toggle_answer(question, answer)
                self.interface.display_capture(self.exam.get_image_drawn())
                self.interface.update_status(
                    self.exam.score,
                    self.exam.decisions.model,
                    self.exam.exam_id,
                    survey_mode=self.exam_data.survey_mode,
                )
                self.session.update_answer(
                    self.exam.exam_id,
                    question,
                    self.exam.capture,
                    self.exam.decisions,
                    self.exam.score,
                    store_captures=False,
                )
                self.interface.run_later(self._store_capture_and_update, delay=100)

    def _mouse_pressed_manual_detection(self, point):
        manager = self.manual_detect_manager
        success = False
        manager.add_point(point)
        self.exam.draw_corner(point)
        self.interface.display_capture(self.exam.get_image_drawn())
        if manager.is_ready():
            success = manager.detect()
            if success:
                new_exam = self._process_capture(manager.detector)
                if new_exam is not None:
                    new_exam.draw_answers()
                else:
                    success = False
            # Remove the exam that was saved previously,
            # before having started the manual review mode:
            self.session.remove_exam(self.exam.exam_id)
            self.interface.remove_exam(self.exam)
            if not success:
                self.exam.reset_image()
                self.interface.show_error(_("Manual detection failed"))
                self._start_search_mode()
            else:
                self.exam = new_exam
                self.from_manual_detection = True
                self._start_review_mode()

    def _exam_selected(self, exam):
        if self.mode.in_grading():
            if self.mode.in_review_from_grading():
                self.exam_id += 1
            self._activate_session_mode()
        exam.load_capture()
        exam.reset_image()
        exam.draw_answers()
        self.exam = exam
        self._start_review_mode()

    def _start_session(self):
        """Starts a session (either a new one or one that has been loaded)."""
        self.interface.add_exams(self.session.read_exams())
        self._activate_session_mode()

    def _activate_session_mode(self):
        self.mode.enter_session()
        self.interface.activate_session_mode()

    def _start_grading(self):
        exam_data = self.exam_data
        self.detection_options = detection.ExamDetector.get_default_options()
        if self.exam_data.survey_mode:
            self.detection_options["infobits"] = False
        self.detection_options["error-logging"] = self.config["error-logging"]
        if exam_data.id_num_digits and exam_data.id_num_digits > 0:
            self.detection_options["read-id"] = True
            self.detection_options["id-num-digits"] = exam_data.id_num_digits
        self.detection_options["left-to-right-numbering"] = (
            exam_data.left_to_right_numbering
        )
        # Set the debug options in detection_options:
        self._action_debug_changed()
        self.detection_context.open_camera()
        if self.detection_context.camera is None:
            self.interface.show_error(
                _("No camera found. Connect a camera and " "start the session again.")
            )
            return
        self.exam_id = self.session.next_exam_id()
        self.interface.clear_selected_exam()
        self._start_search_mode()

    def _stop_grading(self):
        if self.mode.in_grading():
            self.detection_context.close_camera()
        self._activate_session_mode()

    def _store_capture_and_add(self):
        self._store_capture(self.exam)
        self.interface.add_exam(self.exam)
        # Now that the exam is fully processed, we can enter review mode:
        self.mode.enter_review()
        self.interface.activate_review_mode(True)

    def _store_capture_and_update(self):
        self._store_capture(self.exam)
        self.interface.update_exam(self.exam)

    def _store_exam(self, exam):
        self.session.store_exam(
            exam.exam_id, exam.capture, exam.decisions, exam.score, store_captures=False
        )
        self.session.save_raw_capture(exam.exam_id, exam.capture)

    def _store_capture(self, exam):
        self.session.save_drawn_capture(
            exam.exam_id, exam.capture, exam.decisions.student
        )

    def _register_listeners(self):
        listeners = {
            ("actions", "session", "new"): self._new_session,
            ("actions", "session", "open"): self._open_session,
            ("actions", "session", "close"): self._close_session,
            ("actions", "grading", "start"): self._action_start,
            ("actions", "grading", "stop"): self._action_stop,
            ("actions", "grading", "back"): self._action_back,
            ("actions", "grading", "snapshot"): self._action_snapshot,
            ("actions", "grading", "discard"): self._action_discard,
            ("actions", "grading", "continue"): self._action_continue,
            ("actions", "grading", "manual_detect"): self._action_manual_detect,
            ("actions", "grading", "edit_id"): self._action_edit_id,
            ("actions", "tools", "camera"): self._action_camera_selection,
            ("actions", "tools", "export_exam_config"): self._action_export_exam_config,
            ("actions", "tools", "edit_scores"): self._action_edit_scores,
            ("actions", "tools", "lines"): self._action_debug_changed,
            ("actions", "tools", "processed"): self._action_debug_changed,
            ("actions", "tools", "show_status"): self._action_debug_changed,
            ("actions", "tools", "auto_change"): self._action_auto_change_changed,
            ("actions", "exams", "export"): self._action_export_grades,
            ("actions", "exams", "statistics"): self._action_show_statistics,
            ("actions", "exams", "students"): self._action_students,
            ("actions", "help", "help"): self._action_help,
            ("actions", "help", "website"): self._action_website,
            ("actions", "help", "source"): self._action_source_code,
            ("center_view", "camview", "mouse_pressed"): self._mouse_pressed,
            ("window", "exit"): self._exit_application,
            ("window", "exam", "selected"): self._exam_selected,
        }
        self.interface.register_listeners(listeners)


def main():
    # For the translations to work, the initialization of QApplication and
    # the loading of the translations must be done here instead of the
    # gui module:
    #
    from PyQt6.QtWidgets import QApplication
    from PyQt6.QtCore import QTranslator, QLocale, QLibraryInfo

    app = QApplication(sys.argv)
    translator = QTranslator()
    success = translator.load(
        QLocale.system(),
        "qt",
        "_",
        QLibraryInfo.path(QLibraryInfo.LibraryPath.TranslationsPath),
    )
    if not success:
        success = translator.load(
            QLocale.system(), "qt", "_", utils.qt_translations_dir()
        )
    app.installTranslator(translator)
    if len(sys.argv) >= 2:
        filename = sys.argv[1]
    else:
        filename = None
    try:
        interface = gui.Interface(
            app, False, False, preferred_styles=utils.config["gui-styles"]
        )
        manager = ProgramManager(interface, session_file=filename)
        manager.run()
    except utils.EyegradeException as ex:
        print(ex)
        sys.exit(1)


if __name__ == "__main__":
    main()
